use std;

pub trait Q where Self: Sized {
	fn from_raw(i32) -> Self;
	fn qpos() -> usize;
	fn raw(&self) -> i32;

#[cfg_attr(not(release), inline(never))]
	fn one() -> i32 {
		Self::from_i32(1).raw()
	}
#[cfg_attr(not(release), inline(never))]
	fn from_i32(x: i32) -> Self {
		Self::from_raw(x << Self::qpos())
	}
#[cfg_attr(not(release), inline(never))]
	fn muli32(self, x: i32) -> i32 {
		(self.raw() * x) >> Self::qpos()
	}
#[cfg_attr(not(release), inline(never))]
	fn to_i32(self) -> i32 {
		self.raw() >> Self::qpos()
	}
#[cfg_attr(not(release), inline(never))]
	fn raw_i16(self) -> i16 {
		self.raw() as i16
	}
#[cfg_attr(not(release), inline(never))]
	fn from_q<T>(q: T) -> Self where T: Q {
		if T::qpos() < Self::qpos() {
			Self::from_raw(q.raw() << (Self::qpos() - T::qpos()))
		} else {
			Self::from_raw(q.raw() >> (T::qpos() - Self::qpos()))
		}
	}
}

// use macro to make each structs are different type each other.
macro_rules! def_q {
	(pub $name:ident, $qpos:expr) => {
#[derive(Copy, Clone, Debug, Default, Eq, PartialEq)]
		pub struct $name {
			x: i32, // sq(31-qpos).qpos
		}
		impl $crate::fastmath::Q for $name { // FIXME saying self module name is ugly...
			fn from_raw(x: i32) -> Self {
				$name { x: x }
			}
			fn qpos() -> usize { $qpos }
			fn raw(&self) -> i32 { self.x }
		}
		impl std::ops::Add for $name {
			type Output = Self;
			fn add(self, other: Self) -> Self {
				$name { x: self.raw() + other.raw() }
			}
		}
		impl std::ops::AddAssign for $name {
			fn add_assign(&mut self, other: Self) {
				self.x += other.raw();
			}
		}
		impl std::ops::Sub for $name {
			type Output = Self;
			fn sub(self, other: Self) -> Self {
				$name { x: self.raw() - other.raw() }
			}
		}
		impl std::cmp::PartialOrd for $name {
			fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
				self.x.partial_cmp(&other.raw())
			}
		}
		impl std::cmp::Ord for $name {
			fn cmp(&self, other: &Self) -> std::cmp::Ordering {
				self.x.cmp(&other.raw())
			}
		}
	}
}

def_q!(pub Qang, 12); // brad(binary-radian)
def_q!(pub Qsca, 12);

impl Qang {
#[cfg_attr(not(release), inline(never))]
	pub fn from_deg(deg: i32) -> Qang {
		Qang::from_raw(deg * Qang::one() / 360)
	}
}

// 3rd-order: http://www.coranac.com/2009/07/sines/
#[cfg_attr(not(release), inline(never))]
pub fn sin(x: Qang) -> Qsca {
	// S(x) = x * ( (3 << p) - (x * x >> r) ) >> s
	// n: Q-pos for "quarter" circle (2*15 units/circle)
	// A: Q-pos for output
	// p: Q-pos for paren interm
	// r = 2n-p
	// s = A-1-p-n
	let qn = Qang::qpos() - 2;
	let qa = Qsca::qpos();
	let qp = 15; // Qsca::qpos()+3?? 32-Qang::qpos()-2??
	let qr = 2 * qn - qp;
	let qs = qn + qp + 1 - qa;
	// Q13->Q30
	let x1 = i32::wrapping_shl(x.raw(), (30 - qn) as u32); // why u32 not usize?
	// test for quadrant 1/2
	let x2 = if x1 ^ i32::wrapping_shl(x1, 1) < 0 { i32::wrapping_sub(i32::wrapping_shl(1, 31), x1) } else { x1 };
	// Q30->Q13
	let x3 = x2 >> (30 - qn);
	return Qsca::from_raw((x3 * ((3 << qp) - ((x3 * x3) >> qr))) >> qs);
}

#[cfg_attr(not(release), inline(never))]
pub fn cos(x: Qang) -> Qsca {
	sin(x + Qang::from_deg(90))
}

// http://dspguru.com/dsp/tricks/fixed-point-atan2-with-self-normalization
#[cfg_attr(not(release), inline(never))]
pub fn atan2(qy: Qsca, qx: Qsca) -> Qang {
	debug_assert!(Qsca::qpos() <= Qang::qpos());
	let x = qx.raw();
	let y = qy.raw();
	let coeff_1 = ((1 << Qang::qpos()) / 2) / 4; // 3.1415926 / 4;
	let coeff_2 = 3 * coeff_1;
	let abs_y = y.abs() + 1; // add epsilon for avoiding 0/0 condition
	// TODO overflowing at r dividee...
	let (r, asubee) = if 0 <= x {
		let r = ((x - abs_y) << Qang::qpos()) / (x + abs_y);
		(r, coeff_1)
	} else {
		// note: divider swaps!
		let r = ((x + abs_y) << Qang::qpos()) / (abs_y - x);
		(r, coeff_2)
	};
	let angle = asubee - ((coeff_1 * r) >> Qang::qpos());
	return Qang::from_raw(if 0 <= y { angle } else { -angle });
}
