use std;
use std::intrinsics::{volatile_load, volatile_store};
use std::cmp::{min, max};
use ::{gp0, gp1, binread, gpu_byteload, Scene, gpu_setres_raw, gpu_opq_raw15_quad, fillvram, spu_dmaload, spu_play, spucnt_unmute, spu_vol, spucnt_reverb, set_draw_area, cache_ascii_font, draw_ascii_text_cached, gpu_opq_mono_line, SPUMemAllocator};
use ::sys;
use ::fastmath::{sin, cos, atan2, Q, Qsca, Qang};
use ::prng::randn;
use ::progress::Progress;

def_q!(pub Qxy, 8);

#[derive(Default)]
pub struct FlappyScene {
	frame: u32,
	btnstate: sys::DigitalPadButtonState,
	dokans: Vec<FlappyDokan>,
	flappyx: i32,
	flappyy: Qxy,
	flappyvy: Qxy,
	//vb_hassin: Box<[u8]>,
	//vb_gameover: Box<[u8]>,
	va_jump: u32,
	va_crash: u32,
	//revlrvol: (u16, u16),
	died: bool,
	score: u32,
	hiscore: u32,
	hidden: u32,
}

struct FlappyDokan {
	x: i32,
	ystart: i32,
	height: u32,
}

impl FlappyScene {
#[cfg_attr(not(release), inline(never))]
	pub fn new() -> FlappyScene {
		let mut progress = Progress::new(3);
		let mut spu = SPUMemAllocator::new();

		cache_ascii_font(0xFFFF); // explicit.

		// FIXME Load texture before sound: XEBRA BIOS emits noise, NO$PSX BIOS load stops while loading tex.
		//logger.writeln(b"Loading TEX_15.RAW");
		let texbyte = binread(b"cdrom:\\TEX_15.RAW;1\0").expect("read TEX_15.RAW failed");
		//logger.writeln(b"gpuload TEX_15...");
		gpu_byteload(640, 0, 256, 256, &texbyte);
		progress.step();

		let va_jump = spu.load_file(b"cdrom:\\JUMP.VB;1\0");
		progress.step();
		let va_crash = spu.load_file(b"cdrom:\\CRASH.VB;1\0");
		progress.step();
		//logger.writeln(b"Loading HASSIN.VB");
		//let vb_hassin = vec![].into_boxed_slice();//binread(b"cdrom:\\HASSIN.VB\0").expect("could not read HASSIN.VB");
		//logger.writeln(b"Loading GAMEOVER.VB");
		//let vb_gameover = vec![].into_boxed_slice();//binread(b"cdrom:\\GAMEOVER.VB\0").expect("could not read GAMEOVER.VB");

		// Space Echo
		unsafe {
			let params: [u16; 32] = [
				0x033D,0x0231,0x7E00,0x5000,0xB400,0xB000,0x4C00,0xB000,
				0x6000,0x5400,0x1ED6,0x1A31,0x1D14,0x183B,0x1BC2,0x16B2,
				0x1A32,0x15EF,0x15EE,0x1055,0x1334,0x0F2D,0x11F6,0x0C5D,
				0x1056,0x0AE1,0x0AE0,0x07A2,0x0464,0x0232,0x8000,0x8000
			];
			volatile_store(0x1F801DA2 as *mut u16, 0xE128);
			for (i,p) in params.iter().enumerate() {
				volatile_store((0x1F801DC0 + i * 2) as *mut u16, *p);
			}
		}
		spucnt_unmute(true);
		spucnt_reverb(true);
		spu_vol((0x3FFF, 0x3FFF), (0x5EBC, 0x5EBC));

		gpu_setres_raw(3, 0, 0, 0);

		//logger.writeln(b"C-ting Scene");
		FlappyScene {
			//vb_hassin: vb_hassin,
			//vb_gameover: vb_gameover,
			va_jump: va_jump,
			va_crash: va_crash,
			flappyx: 100,
			flappyy: Qxy::from_i32(120),
			flappyvy: Qxy::from_i32(0),
			..Default::default()
		}
	}
}

impl Scene for FlappyScene {
#[cfg_attr(not(release), inline(never))]
	fn draw(&self) {
		gp0(0xE1000000); // texpage (almost for non-texturedpolys)
		gp0(0xE2000000); // texwindow
		gp0(0xE6000000); // maskbits

		// switch displaying frame
		let preyoff = self.frame * 256;
		gp1(0x05000000 | (  0 << 0) | (preyoff << 10));
		// switch drawing frame
		let curyoff = (1 - self.frame) * 256;
		set_draw_area((0, curyoff), (640-1, curyoff + 240-1));
		// fillvram is not affected by drawoffset...
		fillvram(0, 32, 88, 0, curyoff as u16, 640, 240);

		let center = (self.flappyx, self.flappyy.to_i32());
		// instead dy+sqrt, use atan2+sincos.
		let flappy_vx = Qxy::from_i32(1);
		let ang = atan2(Qsca::from_q(self.flappyvy), Qsca::from_q(flappy_vx));
		let (c, s) = (cos(ang).raw(), sin(ang).raw());
		let (dx, dy) = (16, 16);
		let rotxi = |x,y| (x * c - y * s) >> Qang::qpos();
		let rotyi = |x,y| (x * s + y * c) >> Qang::qpos();
		let addvec2i16 = |a:(i32,i32),b:(i32,i32)| ((a.0 + b.0) as i16, (a.1 + b.1) as i16);
		let v1 = addvec2i16(center, (rotxi(-dx,-dy) * 2, rotyi(-dx,-dy)));
		let v2 = addvec2i16(center, (rotxi( dx,-dy) * 2, rotyi( dx,-dy)));
		let v3 = addvec2i16(center, (rotxi(-dx, dy) * 2, rotyi(-dx, dy)));
		let v4 = addvec2i16(center, (rotxi( dx, dy) * 2, rotyi( dx, dy)));
		gpu_opq_raw15_quad(v1, v2, v3, v4, (640 >> 6, 0), (0, 127), (127, 127), (0, 255), (127, 255));

		// 20, 26
		{
			let (dx, dy) = (32, 13);
			for dokan in &self.dokans {
				if 13 < dokan.ystart {
					gpu_opq_raw15_quad(
						addvec2i16((dokan.x, 0), (0, 0)),
						addvec2i16((dokan.x, 0), (dx, 0)),
						addvec2i16((dokan.x, 0), (0, dokan.ystart - 13)),
						addvec2i16((dokan.x, 0), (dx, dokan.ystart - 13)),
						(640 >> 6, 0),
						(128, 0), (255, 0), (128, 20), (255, 20)
						);
				}
				gpu_opq_raw15_quad(
					addvec2i16((dokan.x, dokan.ystart - dy), (0, 0)),
					addvec2i16((dokan.x, dokan.ystart - dy), (dx, 0)),
					addvec2i16((dokan.x, dokan.ystart - dy), (0, dy)),
					addvec2i16((dokan.x, dokan.ystart - dy), (dx, dy)),
					(640 >> 6, 0),
					(128, 20), (255, 20), (128, 46), (255, 46)
					);

				gpu_opq_raw15_quad(
					addvec2i16((dokan.x, dokan.ystart + dokan.height as i32), (0, 0)),
					addvec2i16((dokan.x, dokan.ystart + dokan.height as i32), (dx, 0)),
					addvec2i16((dokan.x, dokan.ystart + dokan.height as i32), (0, dy)),
					addvec2i16((dokan.x, dokan.ystart + dokan.height as i32), (dx, dy)),
					(640 >> 6, 0),
					(128, 46), (255, 46), (128, 20), (255, 20)
					);
				gpu_opq_raw15_quad(
					addvec2i16((dokan.x, dokan.ystart + dokan.height as i32 + dy), (0, 0)),
					addvec2i16((dokan.x, dokan.ystart + dokan.height as i32 + dy), (dx, 0)),
					addvec2i16((dokan.x, 240), (0, 0)),
					addvec2i16((dokan.x, 240), (dx, 0)),
					(640 >> 6, 0),
					(128, 20), (255, 20), (128, 0), (255, 0)
					);
			}
		}

		draw_ascii_text_cached(640 - 32 * 4 - 16, 16, 0x40000, 0x20000, format!("{:4}", self.score).as_bytes());
		gpu_opq_mono_line((255, 255, 255), (640 - 32 * 4 - 16, 33), (640, 33));
		draw_ascii_text_cached(640 - 32 * 4 - 16, 36, 0x40000, 0x20000, format!("{:4}", self.hiscore).as_bytes());
	}
#[cfg_attr(not(release), inline(never))]
	fn input(&mut self, bits: u16) -> Option<Box<Scene>> {
		self.btnstate.update(bits);

		/*
		if self.btnstate.down().triangle() {
			spu_dmaload(0x1000, &self.vb_hassin);
			spu_play(0, 0x1000, 18900);
		}
		if self.btnstate.down().circle() {
			spu_dmaload(0x1000, &self.vb_gameover);
			spu_play(0, 0x1000, 18900);
		}
		*/

		/*
		if self.btnstate.down().square() {
			let pl = 0x1F801D84 as *mut u16;
			let pr = 0x1F801D86 as *mut u16;
			let (vl, vr) = unsafe { (volatile_load(pl), volatile_load(pr)) };
			if vl != 0 {
				self.revlrvol = (vl, vr);
				unsafe {
					volatile_store(pl, 0);
					volatile_store(pr, 0);
				}
			} else {
				unsafe {
					volatile_store(pl, self.revlrvol.0);
					volatile_store(pr, self.revlrvol.1);
				}
			}
		}
		*/

		if !self.died {
			if self.btnstate.down().up() {
				self.flappyvy = Qxy::from_i32(-2);
				spu_play(0, self.va_jump, 22050);
			}

			// move flappyfish
			self.flappyy = max(Qxy::from_i32(0), min(Qxy::from_i32(240), self.flappyy + self.flappyvy));
			self.flappyvy += Qxy::from_raw(10); // 1/256px unit
			// make fresh dokan
			if self.dokans.iter().max_by_key(|e|e.x).and_then(|e|Option::Some(e.x < 400)).or(Option::Some(true)).unwrap() {
				let h = 50 + randn(50);
				self.dokans.push(FlappyDokan {
					x: 640,
					ystart: 25 + randn(240 - 25*2 - h) as i32,
					height: h
				});
			}
			// move dokans left
			for dokan in &mut self.dokans {
				dokan.x -= 2;
			}
			// garbage dokans collect
			self.dokans.retain(|d| -50 < d.x);
			// hittest
			for dokan in &self.dokans {
				if dokan.x < self.flappyx + 16 && // flappy is started to overwrap,
						self.flappyx - 16 < dokan.x + 16 && // and not gone,
						(self.flappyy.to_i32() - 8 < dokan.ystart as i32 || (dokan.ystart + dokan.height as i32) < self.flappyy.to_i32() + 8) {
					self.died = true;
					spu_play(0, self.va_crash, 11025);
					break;
				}
				if dokan.x + 16 == self.flappyx - 16 { // FIXME so fragile test...
					self.score = min(self.score + 1, 9999); // counter-stop
					self.hiscore = max(self.hiscore, self.score);
				}
			}
		} else {
			// died: restart
			if self.btnstate.down().down() {
				self.died = false;
				self.flappyy = Qxy::from_i32(120);
				self.flappyvy = Qxy::from_i32(0);
				self.dokans.clear();
				self.score = 0;
				self.hidden = 0;
			}
			if self.btnstate.down().select() {
				self.hidden += 1;
				if 15 <= self.hidden {
					return Some(Box::new(::titlescene::TitleScene::new()));
				}
			}
		}

		// draw is immutable ref... modify this here.
		self.frame = (self.frame + 1) % 2;

		return None;
	}
}
