use ::{dma, dmawait, binread};

pub struct TDFace {
	pub rgbc: u32,
	pub vis: (usize, usize, usize),
	pub normal: (u32, u16), // GTE optimized
}
pub struct TDModel {
	pub faces: Vec<TDFace>,
	pub vertices: Vec<(u32, u16)>, // GTE optimized
}

impl TDModel {
	#[cfg_attr(not(release), inline(never))]
	pub fn parse<T: AsRef<[u8]>>(data_r: T) -> TDModel {
		let data = data_r.as_ref();
		let vertoff = unsafe { *((data.as_ptr() as *const u32).offset(3)) } as usize + 12;
		let vertnel = unsafe { *((data.as_ptr() as *const u32).offset(4)) } as usize;
		let normoff = unsafe { *((data.as_ptr() as *const u32).offset(5)) } as usize + 12;
		let normnel = unsafe { *((data.as_ptr() as *const u32).offset(6)) } as usize; //optional
		let faceoff = unsafe { *((data.as_ptr() as *const u32).offset(7)) } as usize + 12;
		let facenel = unsafe { *((data.as_ptr() as *const u32).offset(8)) } as usize;
		let getvert = |i| -> (u32, u32) {
			assert!(i < vertnel);
			let p = unsafe { data.as_ptr().offset((vertoff + i * 8) as isize) as *const u32 };
			unsafe { (*p, *p.offset(1)) }
		};
		let getnorm = |i| -> (u32, u32) {
			assert!(i < normnel);
			let p = unsafe { data.as_ptr().offset((normoff + i * 8) as isize) as *const u32 };
			unsafe { (*p, *p.offset(1)) }
		};
		let getface = |i| -> (u32, usize, (usize, usize, usize)) {
			assert!(i < facenel);
			let p = unsafe { data.as_ptr().offset((faceoff as usize + i * 16) as isize) };
			unsafe { (
					*(p.offset(4) as *const u32),
					*(p.offset(8) as *const u16) as usize,
					(
						*(p.offset(10) as *const u16) as usize,
						*(p.offset(12) as *const u16) as usize,
						*(p.offset(14) as *const u16) as usize)) }
		};
		TDModel {
			faces: (0..facenel).map(|i|{
				let f = getface(i);
				let (nxy, nz) = getnorm(f.1);
				TDFace {
					rgbc: f.0,
					normal: (nxy, nz as u16),
					vis: f.2,
				}
			}).collect(),
			vertices: (0..vertnel).map(|i|{
				let (xy, z) = getvert(i);
				(xy, z as u16)
			}).collect(),
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn load(path: &[u8]) -> TDModel {
		let data = match binread(path) {
			Ok(v) => v,
			Err(e) => panic!("Can't load {}: {:?}", ::std::str::from_utf8(path).unwrap(), e)
		};
		Self::parse(&data)
	}
}

pub struct OrderTabDiag {
	pub npoly: usize,
	pub zrange: (u16, u16),
}

pub struct OrderTab {
	ot: Vec<u32>,
	depth: usize,
	first_ptr: u32, // u32 not *mut u32 nor usize, just for convenient.
	diag: OrderTabDiag,
}

impl OrderTab {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(depth: usize, buffer: usize) -> OrderTab {
		let mut ot = Vec::with_capacity(depth + buffer); // FIXME reuse ot? but draw gets &self not &mut self...
		unsafe { ot.set_len(depth); } // initialized by just next dma.
		// note: dma6 initializes OT backward. be careful for ptr(point last element) and ctl(direction=backward).
		dma(6, unsafe { ot.as_mut_ptr().offset(ot.len() as isize - 1) } as *const u8, 0, ot.len() as u16, 0x11000002); // FIXME semantic mismatch: ptr=*mut u8, length=0 blocksize=words
		OrderTab {
			first_ptr: ot.as_mut_ptr() as u32, // keep OT ptr here for relocation of "ot"
			ot: ot,
			depth: depth,
			diag: OrderTabDiag {
				npoly: 0,
				zrange: (0xFFFF, 0x0000),
			},
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn settle(&self) {
		dmawait(6);
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn insert<T: AsRef<[u32]>>(&mut self, idx: usize, cmds_r: T) {
		let cmds = cmds_r.as_ref();
		debug_assert!(idx < self.depth);
		let next = self.ot[idx];
		self.ot[idx] = (self.first_ptr + self.ot.len() as u32 * 4) & 0x00FFffff; // physaddr of end of original OT
		self.ot.push(((cmds.len() as u32) << 24) | next);
		self.ot.extend(cmds);
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn enqueue(&mut self, model: &TDModel) {
		// xlate each vertices and build OT
		for face in &model.faces {
			let v0 = model.vertices[face.vis.0];
			let v1 = model.vertices[face.vis.1];
			let v2 = model.vertices[face.vis.2];
			// coord. xlat. and culling
			mtc2!(0, v0.0); // VXY0
			mtc2!(1, v0.1 as u32); // VZ0
			mtc2!(2, v1.0); // VXY0
			mtc2!(3, v1.1 as u32); // VZ0
			mtc2!(4, v2.0); // VXY0
			mtc2!(5, v2.1 as u32); // VZ0
			rtpt!(1); // V012,TR,H,OFXY,DQAB -> S012,IR0(,MAC0123,IR123)
			nclip!(); // SXY012 -> MAC0
			let nclip_res = mfc2!(24) as i32;
			if nclip_res < 0 {
				// clipping
				continue;
			}

			{
				let (z3, z2, z1) = (mfc2!(19) as u16, mfc2!(18) as u16, mfc2!(17) as u16);
					self.diag.zrange = (
						*[self.diag.zrange.0, z3, z2, z1].iter().min().unwrap(),
						*[self.diag.zrange.1, z3, z2, z1].iter().max().unwrap()
						);
			}

			// calc. Z and clip too near or too far
			avsz3!(); // SZ123,ZSF3 -> OTZ(,MAC0)
			let otz = mfc2!(7) as usize; // u16
			if otz == 0 || self.depth <= otz {
				continue;
			}

			// lighting with depth queing
			mtc2!(6, face.rgbc); // RGBC. note: including cmd 0x20 (mono tri poly opaque)
			mtc2!(0, face.normal.0); // VXY0
			mtc2!(1, face.normal.1 as u32); // VZ0
			nccs!(1); // LL,V0,BK,LC/*,RGBC,FC,IR0*/ -> RGB2(,MAC123,IR0123)

			// enqueue into OT
			let cmd = mfc2!(22) as u32; // RGB2
			let sxy0 = mfc2!(12) as u32;
			let sxy1 = mfc2!(13) as u32;
			let sxy2 = mfc2!(14) as u32;
			self.insert(otz, &[cmd, sxy0, sxy1, sxy2]);
			{
				self.diag.npoly += 1;
			}
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn submit(&mut self) {
		// adjust "ot" if it is relocated
		if self.ot.as_ptr() as u32 != self.first_ptr {
			// FIXME DEBUG keep original(before-reloc) table for inspecting crash
			//let orig_ot = self.ot.clone();
			// diff: Sub<pointer> is undefined in Rust 1.12, and want byte diff here, not word.
			let diff = self.ot.as_ptr() as u32 - self.first_ptr;
			unsafe {
				let mut ptr = self.ot.as_mut_ptr().offset(self.depth as isize - 1); // this first is in kseg0, but next is useg.
				while *ptr != 0x00FFffff {
					*ptr += diff;
					ptr = (*ptr & 0x00FFffff) as *mut u32;
					// FIXME DEBUG
					/*
					let psp = self.ot.as_ptr() as u32 & 0xFFffff;
					let pptr = ptr as u32;
					let pep = psp + self.ot.len() as u32 * 4;
					let over_start = psp <= pptr;
					let before_end = pptr < pep;
					assert!(over_start && before_end, "COSMIC RAY: NOT {:X} <= {:X} < {:X} n={:08X} o={:08X}", psp, pptr, pep, self.ot.as_ptr() as u32, orig_ot.as_ptr() as u32);
					*/
				}
			}
		}

		// xfer OT to GPU
		unsafe { *(0x1F801814 as *mut u32) = 0x04000002; } // GPU DMA direction: CPU->GP0
		dma(2, unsafe { self.ot.as_ptr().offset(self.depth as isize - 1) } as *const u8, 0, 0, 0x01000401); // FIXME semantic mismatch: length/blocksize=dontcare
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn wait(&self) {
		dmawait(2);
	}

	#[allow(dead_code)]
	#[cfg_attr(not(release), inline(never))]
	pub fn get_diag(&self) -> Option<&OrderTabDiag> {
		Option::Some(&self.diag) // TODO None by release cfg
	}
}
