use super::{Scene, GraphicLogger, gpu_setres_raw, gp0, gp1, fillvram, dma, dmawait, draw_ascii_text_cached};
use super::sys::{cd_read, DigitalPadButtonState, DigitalPadButtons};
use std::default::Default;

#[derive(Default)]
pub struct PSLogoScene {
	frame: u32,
	data: Vec<u8>,
	ot_depth: usize,
	trz: u32,
	btns: DigitalPadButtonState,
	btnrepeat: u32,
	fmin: u32,
	fmax: u32,
	poly: u32,
}

impl PSLogoScene {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(_logger: &mut GraphicLogger) -> PSLogoScene {
		gpu_setres_raw(1, 0, 0, 0);
		PSLogoScene {
			data: cd_read(5, 7).expect("could not read PSLogo"),
			ot_depth: 1024,
			trz: 0x1700,
			fmin: 0x200,
			fmax: 0x20C,
			poly: 600,
			..Default::default()
		}
	}
}

impl Scene for PSLogoScene {
	#[cfg_attr(not(release), inline(never))]
	#[allow(overflowing_literals)] // FIXME mtc2
	fn draw(&self) {
		gp0(0xE1000000); // texpage (almost for non-texturedpolys)
		gp0(0xE2000000); // texwindow
		gp0(0xE6000000); // maskbits

		// switch displaying frame
		let preyoff = self.frame * 240;
		gp1(0x05000000 | (  0 << 0) | (preyoff << 10));
		// switch drawing frame
		let curyoff = (1 - self.frame) * 240;
		gp0(0xE3000000 | (  0 << 0) | (curyoff << 10)); // draw(clip)x1y1
		gp0(0xE4000000 | (320 << 0) | ((curyoff + 240) << 10)); // draw(clip)x2y2
		gp0(0xE5000000 | (  0 << 0) | (curyoff << 11)); // drawoffset
		// fillvram is not affected by drawoffset...
		fillvram(64, 64, 64, 0, curyoff as u16, 320, 240);

		assert!(self.data[0] == 0x41);

		// initialize OT (backward)
		let mut ot = Vec::with_capacity(self.ot_depth + 4096/*just guess*/); // FIXME reuse ot? but draw gets &self not &mut self...
		unsafe { ot.set_len(self.ot_depth); }
		dma(6, unsafe { ot.as_ptr().offset(ot.len() as isize - 1) } as *const u8, 0, ot.len() as u16, 0x11000002); // FIXME semantic mismatch: ptr=*mut u8, length=0 blocksize=words

		// matrix and vectors
		mtc2!(32, 0x00000FFB); mtc2!(33, 0xFFB7FF44); mtc2!(34, 0xF9CA0EBC); mtc2!(35, 0x063700AD); mtc2!(36, 0x00000EB7); // RT3x3q12
		//mtc2!(32, 0x00001000); mtc2!(33, 0x00000000); mtc2!(34, 0x00001000); mtc2!(35, 0x00000000); mtc2!(36, 0x00001000); // RT3x3q12
		mtc2!(40, 0x00000000); mtc2!(41, 0x00000FA0); mtc2!(42, 0x0000F060); mtc2!(43, 0x0000F060); mtc2!(44, 0x00000000); // LM3x3q12
		mtc2!(48, 0x0BB80FA0); mtc2!(49, 0x0FA00FA0); mtc2!(50, 0x0FA00BB8); mtc2!(51, 0x0BB80FA0); mtc2!(52, 0x00000FA0); // LL3x3q12
		mtc2!(37, 0x00000000); mtc2!(38, 0x00000000); mtc2!(39, self.trz/*0x00001700*/); // TRq0 0,-170,5888
		mtc2!(45, 0x00000640); mtc2!(46, 0x00000640); mtc2!(47, 0x00000640); // BKq12?
		mtc2!(53, 0x00000000); mtc2!(54, 0x00000000); mtc2!(55, 0x00000000); // FCq4
		// for RTP
		mtc2!(56, 0x00A00000); mtc2!(57, 0x00780000); // OFX,OFYq16 160,120
		mtc2!(58, 0x0400); // Hq0u16
		mtc2!(59, 0xFC82); // DQAq8 (coeff)
		mtc2!(60, 0x01400000); // DQBq24? (offset)
		// for AVSZ
		mtc2!(61, 0x0155>>1); // ZSF3q12 (1/3 / 2)
		mtc2!(62, 0x0100>>1); // ZSF4q12 (1/4 / 2)

		let first_ot_ptr = ot.as_ptr(); // keep here OT ptr for relocation of "ot"
		let mut npoly = 0;
		// xlate each vertices and build OT
		{
			let vertoff = unsafe { *((self.data.as_ptr() as *const u32).offset(3)) } as usize + 12;
			let vertnel = unsafe { *((self.data.as_ptr() as *const u32).offset(4)) } as usize; //optional
			let normoff = unsafe { *((self.data.as_ptr() as *const u32).offset(5)) } as usize + 12;
			let normnel = unsafe { *((self.data.as_ptr() as *const u32).offset(6)) } as usize; //optional
			let faceoff = unsafe { *((self.data.as_ptr() as *const u32).offset(7)) } as usize + 12;
			let facenel = unsafe { *((self.data.as_ptr() as *const u32).offset(8)) } as usize;
			let getvert = |i| -> (u32, u32) {
				assert!(i < vertnel);
				let p = unsafe { self.data.as_ptr().offset((vertoff + i * 8) as isize) as *const u32 };
				unsafe { (*p, *p.offset(1)) }
			};
			let getnorm = |i| -> (u32, u32) {
				assert!(i < normnel);
				let p = unsafe { self.data.as_ptr().offset((normoff + i * 8) as isize) as *const u32 };
				unsafe { (*p, *p.offset(1)) }
			};
			let getface = |i| -> (u32, usize, (usize, usize, usize)) {
				assert!(i < facenel);
				let p = unsafe { self.data.as_ptr().offset((faceoff as usize + i * 16) as isize) };
				unsafe { (
						*(p.offset(4) as *const u32),
						*(p.offset(8) as *const u16) as usize,
						(
							*(p.offset(10) as *const u16) as usize,
							*(p.offset(12) as *const u16) as usize,
							*(p.offset(14) as *const u16) as usize)) }
			};
			let mut ot_insert = |i, cmds: &[u32]| {
				assert!(i < self.ot_depth);
				let next = ot[i];
				ot[i] = unsafe { first_ot_ptr.offset(ot.len() as isize) as u32 } & 0x00FFffff; // physaddr of end of original OT
				ot.push(((cmds.len() as u32) << 24) | next);
				ot.extend(cmds);
			};
			for i in 0..facenel {
				let face = getface(i);
				let rgbc = face.0; // note: including cmd 0x20 (mono tri poly opaque)
				let norm = getnorm(face.1);
				let vis = face.2;
				let v0 = getvert(vis.0);
				let v1 = getvert(vis.1);
				let v2 = getvert(vis.2);
				// coord. xlat. and culling
				mtc2!(0, v0.0); // VXY0
				mtc2!(1, v0.1); // VZ0
				mtc2!(2, v1.0); // VXY0
				mtc2!(3, v1.1); // VZ0
				mtc2!(4, v2.0); // VXY0
				mtc2!(5, v2.1); // VZ0
				rtpt!(1); // V012,TR,H,OFXY,DQAB -> S012,IR0(,MAC0123,IR123)
				nclip!(); // SXY012 -> MAC0
				let nclip_res = mfc2!(24) as i32;
				if nclip_res < 0 {
					// clipping
					continue;
				}

				// calc. Z and clip too near or too far
				avsz3!(); // SZ123,ZSF3 -> OTZ(,MAC0)
				let otz = mfc2!(7) as usize; // u16
				if otz == 0 || self.ot_depth <= otz {
					continue;
				}

				// lighting with depth queing
				mtc2!(6, rgbc); // RGBC
				mtc2!(0, norm.0); // VXY0
				mtc2!(1, norm.1); // VZ0
				ncds!(1); // LL,V0,BK,LC,RGBC,FC,IR0 -> RGB2(,MAC123,IR0123)

				// enqueue into OT
				let cmd = mfc2!(22) as u32; // RGB2
				let sxy0 = mfc2!(12) as u32;
				let sxy1 = mfc2!(13) as u32;
				let sxy2 = mfc2!(14) as u32;
				if npoly < self.poly && (otz < self.fmin as usize || (self.fmax as usize) < otz) {
					ot_insert(otz, &[cmd, sxy0, sxy1, sxy2][..]); // TODO wtf [..] required??
				}
				npoly += 1;
			}
		}

		// adjust "ot" if it is relocated
		if ot.as_ptr() != first_ot_ptr {
			// diff: Sub<pointer> is undefined in Rust 1.12, and want byte diff here, not word.
			let diff = ot.as_ptr() as u32 - first_ot_ptr as u32;
			unsafe {
				let mut ptr = ot.as_mut_ptr().offset(self.ot_depth as isize - 1); // this first is in kseg0, but next is useg.
				while *ptr != 0x00FFffff {
					*ptr += diff;
					ptr = (*ptr & 0x00FFffff) as *mut u32;
				}
			}
		}

		// xfer OT to GPU
		dma(2, unsafe { ot.as_ptr().offset(self.ot_depth as isize - 1) } as *const u8, 0, 0, 0x01000401); // FIXME semantic mismatch: length/blocksize=dontcare
		dmawait(2); // required on XEBRA 150831/ePSXe 2.0.5
		draw_ascii_text_cached(0, 16, 0x10000, 0x10000, &format!("trz={:X}", self.trz).as_bytes());
		draw_ascii_text_cached(0, 24, 0x10000, 0x10000, &format!("fmin={:X}", self.fmin).as_bytes());
		draw_ascii_text_cached(0, 32, 0x10000, 0x10000, &format!("fmax={:X}", self.fmax).as_bytes());
		draw_ascii_text_cached(0, 40, 0x10000, 0x10000, &format!("poly={}/{}", self.poly, npoly).as_bytes());

	}
	#[cfg_attr(not(release), inline(never))]
	fn input(&mut self, bits: u16) {
		self.btns.update(bits);
		let press = self.btns.press();
		self.btnrepeat = if !press.up() && !press.down() { 0 } else { self.btnrepeat + 1 };
		let down = self.btns.down() | if 30 < self.btnrepeat { press } else { DigitalPadButtons::new(0) };

		let mut f =
			if press.triangle() {
				&mut self.fmax
			} else if press.cross() {
				&mut self.fmin
			} else if press.circle() {
				&mut self.poly
			} else {
				&mut self.trz
			};
		if down.up() {
			*f -= if press.square() { 16 } else { 1 };
		}
		if down.down() {
			*f += if press.square() { 16 } else { 1 };
		}

		self.frame = (self.frame + 1) % 2;
	}
}
