use ::{Scene, binread, spu_dmaload, spucnt_unmute, spucnt_reverb, spu_vol, spu_play_ex, SPUMemAllocator};
use ::{gpu_setres_raw, fillvram, gp0, gp1, draw_ascii_text_cached, gpu_monotri, gpu_calc_texpage, SemiTrans, gpu_opq_mono_line, gpu_opq_mono_polyline, gpu_polyline_start, gpu_byteload, gpu_semitrans_mono_rect, gpu_opq_raw_rect, gpu_opq_shaded_polyline, gpu_semitrans_texblend_quad, gpu_opq_raw15_quad, gpu_semitrans_texblend15_quad, gpu_opq_mono_rect, set_draw_area}; // TODO should move to graphic
use ::sys::{DigitalPadButtonState};
use ::fastmath::{Qang, Qsca, Q, sin, cos};
use ::graphic::{TDModel, OrderTab};
use ::mat::{Vec3Q12, Vec3wQ12, load_mat9_rt, load_mat9_ll, load_mat9_lc, mult_load_mat9_rt, mult_load_mat9_ll, mult_load_mat9_lc, ir, get_mat33trgte_rt, load_mat33trgte_rt, cpu_lookat_mat9tr};
use ::prng::rand;
use ::progress::Progress;
use ::std::cmp::{min, max};
use ::std::rc::{Rc, Weak};
use ::std::cell::{Cell, RefCell};

#[derive(Clone, Default)] // Default but use new()!
pub struct Beater {
	bpm: i32,
	onbeat: bool,
	onsubbeat: bool,
	bhus: i32, // beat hus
	nexthus: i32, // hus: hundred micro sec
}

impl Beater {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(bpm: i32) -> Beater {
		let mut me = Beater { ..Default::default() };
		me.set_bpm(bpm);
		me
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn set_bpm(&mut self, bpm: i32) {
		self.bpm = bpm;
		self.bhus = 600000/*1min in hus*/ / self.bpm;
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn update(&mut self) {
		self.onbeat = false;
		self.onsubbeat = false;

		let pb = self.nexthus / (self.bhus / 4);
		self.nexthus -= 166; // 1/60s in hus
		let cb = self.nexthus / (self.bhus / 4);
		if cb != pb {
			self.onsubbeat = true;
		}
		if self.nexthus <= 0 {
			self.nexthus += self.bhus;
			self.onbeat = true;
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn beat_lerp(&self, max: i32) -> i32 {
		max * self.nexthus / self.bhus
	}

	pub fn bpm(&self) -> i32 { self.bpm }
	pub fn onbeat(&self) -> bool { self.onbeat }
	pub fn onsubbeat(&self) -> bool { self.onsubbeat }
}

fn xlate_to_screen((x, y, z): (i16, i16, i16)) -> (u32, u32) {
	mtc2!(0, (x as u16 as u32) | ((y as u16 as u32) << 16)); // VXY0
	mtc2!(1, z as u32); // VZ0
	rtps!(1); // V0,TR,H,OFXY,DQAB -> S2,IR0(,MAC0123,IR123)
	(mfc2!(14) as u32, mfc2!(15) as u32) // SXY2,SZ2
}

#[derive(Default)] // but use new()!
pub struct Peaker {
	peaklevel: u32,
	peakbase: u32,
	widthlevel: u32,
	widthbase: u32,
	noiselevel: u32,

	peaks: [[i32; 25]; 32], // should ringbuf?
	peakz: i32,
	peakbeated: bool,
}

impl Peaker {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(peaklevel: u32, peakbase: u32, widthlevel: u32, widthbase: u32, noiselevel: u32) -> Peaker {
		let mut me = Peaker {
			peaklevel: peaklevel,
			peakbase: peakbase,
			widthlevel: widthlevel,
			widthbase: widthbase,
			noiselevel: noiselevel,
			..Default::default()
		};
		for _ in 0..me.peaks.len() { me.gen(); }
		me
	}

	// draw peak from far to Z~+1000
	#[cfg_attr(not(release), inline(never))]
	pub fn draw(&self) {
		for (z, peaks) in self.peaks.iter().enumerate() {
			let xlated: Vec<(u32,u32)> = peaks.iter().enumerate().map(|(i,p)| {
				let x: i16 = (i as i16 - 12) * 30;
				let y: i16 = *p as i16 * 3;
				let z: i16 = (z as i16) * 450 + self.peakz as i16 - 12000; // TODO does not overflow in i16?
				xlate_to_screen((x, y, z))
			}).collect();

			gpu_polyline_start(0x48, (0xFF, 0, 0)); // 48: mono-poly-opq
			for (w,z) in xlated {
				if 0 < z { // draw only non-near
					gp0(w);
				}
			}
			gp0(0x5555_5555); // magic terminator for polyline
		}
	}

	#[cfg_attr(not(release), inline(never))]
	pub fn update(&mut self, beater: &Beater, speed: i32) {
		self.peakbeated |= beater.onbeat();
		self.peakz += speed;
		if 450 <= self.peakz {
			self.peakz -= 450;
			self.gen();
		}
	}

	#[cfg_attr(not(release), inline(never))]
	fn gen(&mut self) {
		let w = self.peaks[0].len();
		// shift
		for i in (1..self.peaks.len()).rev() {
			self.peaks[i] = self.peaks[i - 1];
		}
		// make new
		for i in 0..w {
			self.peaks[0][i] = self.peaks[1][i] * 9 / 10;
		}
		if self.peakbeated {
			self.peakbeated = false;
			let px = (rand() % 19 + 3) as i32;
			let pw = (rand() % self.widthlevel + self.widthbase) as i32;
			let py = (rand() % self.peaklevel + self.peakbase) as i32;
			for x in 0..w {
				let dx = (px - x as i32).abs();
				let y = if dx < pw { cos(Qang::from_raw(Qang::one() * dx / pw / 2)).muli32(py) } else { 0 };
				let it = &mut self.peaks[0][x as usize];
				*it = max(*it, y);
			}
		}
		for x in 0..w {
			self.peaks[0][x as usize] += (rand() % self.noiselevel) as i32;
		}
	}
}

pub trait Bullet {
	fn pos(&self) -> (i16, i16, i16);
	fn draw(&self, ot: &OrderTab);
	fn update(&mut self);
	fn ishitting(&self) -> bool;
}

pub trait Enemy {
	fn pos(&self) -> (i16, i16, i16);
	fn draw(&self, ot: &mut OrderTab);
	fn update(&mut self, beater: &Beater) -> Vec<Box<Bullet>>;
	fn can_lockon(&self) -> bool;
	fn lockon(&mut self);
	fn hit(&mut self);
	fn alive(&self) -> bool;

	// FIXME should eliminated
	fn cached_spos(&self) -> (i16, i16);
	fn update_spos_cache(&self, sxy: (i16, i16));
}

trait Effect {
	fn pos(&self) -> (i16, i16, i16);
	fn draw_effectarea(&self);
	fn draw_before(&self);
	fn draw_inter(&self, ot: &mut OrderTab);
	fn draw_after(&self);
	fn update(&mut self);
	fn alive(&self) -> bool;
}

/*
struct BulletA {
}

impl Bullet for BulletA {
	fn pos(&self) -> (i32, i32, i32);
	fn draw(&self, ot: &OrderTab);
	fn update(&mut self) -> u32;
	fn ishitting(&self) -> bool;
}
*/

struct EnemyE {
	model: Rc<TDModel>,
	pos: (i16, i16, i16),
	#[allow(dead_code)]
	frame: i32,
	locked: bool,
	died: bool,

	cached_spos: Cell<(i16, i16)>, // FIXME OMG INTERIOR MUTABLE
}

impl EnemyE {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(models: &Models, pos: (i16, i16, i16)) -> EnemyE {
		EnemyE {
			model: models.enemye.clone(),
			pos: pos,
			frame: 0,
			locked: false,
			died: false,

			cached_spos: Cell::new((0, 0)),
		}
	}
}

impl Enemy for EnemyE {
	#[cfg_attr(not(release), inline(never))]
	fn pos(&self) -> (i16, i16, i16) { self.pos }
	#[cfg_attr(not(release), inline(never))]
	fn draw(&self, ot: &mut OrderTab) { mult_load_mat9_rt(&mat_scale1(0x300)); ot.enqueue(&self.model); }
	#[cfg_attr(not(release), inline(never))]
	fn update(&mut self, _beater: &Beater) -> Vec<Box<Bullet>> { vec![] }
	#[cfg_attr(not(release), inline(never))]
	fn can_lockon(&self) -> bool { !self.locked }
	#[cfg_attr(not(release), inline(never))]
	fn lockon(&mut self) {
		self.locked = true;
	}
	#[cfg_attr(not(release), inline(never))]
	fn hit(&mut self) {
		/*HP is 1, die.*/
		self.died = true;
	}
	#[cfg_attr(not(release), inline(never))]
	fn alive(&self) -> bool { !self.died }

	#[cfg_attr(not(release), inline(never))]
	fn cached_spos(&self) -> (i16, i16) { self.cached_spos.get() }
	#[cfg_attr(not(release), inline(never))]
	fn update_spos_cache(&self, sxy: (i16, i16)) { self.cached_spos.set(sxy); }
}

struct SonicCircleEffect {
	pos: (i16, i16, i16),
	radius: i32,
	frame: i32,
}

impl SonicCircleEffect {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(pos: (i16, i16, i16), radius: i32) -> SonicCircleEffect {
		SonicCircleEffect {
			pos: pos,
			radius: radius,
			frame: 0,
		}
	}
}

impl Effect for SonicCircleEffect {
	fn pos(&self) -> (i16, i16, i16) { self.pos }
	fn draw_effectarea(&self) { /*nothing to do*/ }
	#[cfg_attr(not(release), inline(never))]
	fn draw_before(&self) {
		let wrinkletime = 20;
		let fadetime = 15;
		let alltime = wrinkletime + fadetime - 5;
		let lumi = (255 - max(0, self.frame - (alltime - fadetime)) * 255 / fadetime) as u8;
		let draw_circle = |rad: i32| {
			let res = 32;
			gpu_polyline_start(0x48, (lumi, lumi, lumi));
			for i in 0..(res + 1) {
				let x = cos(Qang::from_raw(Qang::one() * i / (res / 2))).muli32(rad) as i16;
				let y = sin(Qang::from_raw(Qang::one() * i / (res / 2))).muli32(rad) as i16;
				let (sxy, sz) = xlate_to_screen((x, y, 0));
				if 0 < sz { // draw only non-near TODO this function should be merged (with peaker.draw)!!
					gp0(sxy);
				}
			}
			gp0(0x55555555); // magic terminate word
		};
		fn coslerp(num: i32, denom: i32) -> Qsca {
			cos(Qang::from_raw(num * Qang::one() / 4 / denom))
		};
		// TODO should think more.
		let rad = coslerp(max(0, (wrinkletime - self.frame)), wrinkletime).muli32(self.radius);
		draw_circle(rad);
		draw_circle(coslerp(alltime - self.frame, alltime).muli32(rad));
	}
	fn draw_inter(&self, _ot: &mut OrderTab) { /*nothing to do*/ }
	fn draw_after(&self) { /*nothing to do*/ }
	#[cfg_attr(not(release), inline(never))]
	fn update(&mut self) {
		self.frame += 1;
	}
	fn alive(&self) -> bool {
		let wrinkletime = 20;
		let fadetime = 15;
		let alltime = wrinkletime + fadetime - 5;
		self.frame < alltime
	}
}

struct SoftBoomEffect {
	pos: (i16, i16, i16),
	frame: i32,
}

impl SoftBoomEffect {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(pos: (i16, i16, i16)) -> SoftBoomEffect {
		SoftBoomEffect {
			pos: pos,
			frame: 0,
		}
	}
}

impl Effect for SoftBoomEffect {
	fn pos(&self) -> (i16, i16, i16) { self.pos }
	#[cfg_attr(not(release), inline(never))]
	fn draw_effectarea(&self) {
		let lumi = 128 - (self.frame * 128 / 15) as u8; // max 0x80 for x1.0
		let rad = 250 + self.frame * 10;
		let t = Qang::from_deg(150 - (self.frame * 60 / 15));
		let trx = cos(t).muli32(rad) as i16;
		let try = sin(t).muli32(rad) as i16;
		fn xlate(x: i16, y: i16) -> (i16, i16) {
			let (sxy, _sz) = xlate_to_screen((x, y, 0));
			((sxy & 0xFFFF) as i16, (sxy >> 16) as i16)
		}
		let v1 = xlate(trx, try);
		let v2 = xlate(try, -trx);
		let v3 = xlate(-try, trx);
		let v4 = xlate(-trx, -try);
		gpu_semitrans_texblend_quad((lumi, lumi, lumi),
			v1, v2, v3, v4,
			(9, 0), (25, 16), (25+32, 16), (25, 16+32), (25+32, 16+32),
			SemiTrans::Add, 2, (0, 0));
	}
	fn draw_before(&self) { /*nothing to do*/ }
	fn draw_inter(&self, _ot: &mut OrderTab) { /*nothing to do*/ }
	fn draw_after(&self) { /*nothing to do*/ }
	#[cfg_attr(not(release), inline(never))]
	fn update(&mut self) {
		self.frame += 1;
	}
	fn alive(&self) -> bool {
		self.frame < 15
	}
}

pub struct Models {
	humanoid: Rc<TDModel>,
	//enemya: Rc<TDModel>,
	//enemyc: Rc<TDModel>,
	//enemyd: Rc<TDModel>,
	enemye: Rc<TDModel>,
	//bulleta: Rc<TDModel>,
	//bulletb: Rc<TDModel>,
}

pub struct Sounds {
	//bassdrum: u32,
	closehat: u32,
	//choir: u32,
	perc1: u32,
	accsnare: u32,
	//handcrap: u32,
}

trait MusicSequencer {
	fn update(&mut self, beater: &mut Beater);
	fn next_rhythm_in(&self, beater: &Beater, lastrid: u32, frames: i32) -> Option<u32>;
	fn lockedon(&mut self);
	fn hit_enemy(&mut self); // really?
	fn current_rhythm(&self) -> u32;
}

pub struct MusicSequencerLayer0 {
	sounds: Rc<Sounds>,
	hitpattern: Vec<bool>,
	currentsub: i32,
	lockedon: bool,
	hitted: bool,
	rid: u32,
}

impl MusicSequencerLayer0 {
	// TODO ugly, but how load?
	#[cfg_attr(not(release), inline(never))]
	pub fn new(sounds: Rc<Sounds>, beater: &mut Beater) -> MusicSequencerLayer0 {
		beater.set_bpm(85);
		MusicSequencerLayer0 {
			sounds: sounds,
			hitpattern: vec![
				true, false, false, false,
				true, true, false, false,
				true, true, false, true,
				false, false, true, false,
				// loop
				true, false, false, false,
				true, true, false, false,
				true, true, false, true,
				false, false, true, false,
			],
			currentsub: 0,
			lockedon: false,
			hitted: false,
			rid: 0,
		}
	}
}

impl MusicSequencer for MusicSequencerLayer0 {
	#[cfg_attr(not(release), inline(never))]
	fn update(&mut self, beater: &mut Beater) {
		/*
		if beater.onbeat() {
			spu_play_ex(0, self.sounds.bassdrum, /*88185@C5/4=*/22046/*@C3*/, (0x3FFF, 0x3FFF), 0x000F5F00, false);
		} else if beater.onsubbeat() {
			spu_play_ex(1, self.sounds.closehat, /*62356@C5/4*1.410=*/21980/*@F#3*/, (0x3FFF, 0x3FFF), 0x000F5F00, false);
		}
		*/
		if beater.onsubbeat() {
			if self.currentsub == 0 {
				spu_play_ex(0, self.sounds.perc1, 22050, (0x3FFF, 0x3FFF), 0x000F5F00, true);
			}
			if self.lockedon {
				spu_play_ex(1, self.sounds.closehat, /*62356@C5/4*1.410=*/21980/*@F#3*/, (0x3FFF, 0x3FFF), 0x000F5F00, false);
				self.lockedon = false;
			}
			// sound here maybe too late.
			/*
			if self.hitted {
				spu_play_ex(1, self.sounds.accsnare, /*78564@C5->C3=/4=*/19641, (0x3FFF, 0x3FFF), 0x000F5F00, false);
				self.hitted = false;
			}
			*/
			if self.hitpattern[self.currentsub as usize] {
				self.rid += 1;
			}
			self.currentsub = (self.currentsub + 1) % self.hitpattern.len() as i32;
		}
	}
	#[cfg_attr(not(release), inline(never))]
	fn next_rhythm_in(&self, beater: &Beater, lastrid: u32, frames: i32) -> Option<u32> {
		// TODO should calculate. emulating Beater is slow and requires Clone for Beater.
		let mut ber = beater.clone(); // FIXME clone!! is this looks good?
		let mut cursub = self.currentsub;
		let mut rid = self.rid;
		for _ in 0..(frames + 1) {
			if ber.onsubbeat() {
				cursub = (cursub + 1) % self.hitpattern.len() as i32;
				if self.hitpattern[cursub as usize] {
					rid += 1;
					if lastrid < rid {
						return Some(rid);
					}
				}
			}
			ber.update();
		}
		return None;
	}
	#[cfg_attr(not(release), inline(never))]
	fn lockedon(&mut self) {
		self.lockedon = true;
	}
	#[cfg_attr(not(release), inline(never))]
	fn hit_enemy(&mut self) {
		self.hitted = true;
	}

	fn current_rhythm(&self) -> u32 { self.rid }
}

trait EnemySequencer {
	fn update(&mut self, beater: &Beater) -> Vec<Box<Enemy>>;
}

// TODO super ugly hard coding!! but I can't write fn array...
pub struct EnemySequencerLayer0 {
	subbeats: u32,
	models: Rc<Models>,
	y: i16,
}

impl EnemySequencerLayer0 {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(models: Rc<Models>) -> EnemySequencerLayer0 {
		EnemySequencerLayer0 {
			subbeats: 0,
			models: models,
			y: (rand() % 400) as i16 - 200,
		}
	}
}
impl EnemySequencer for EnemySequencerLayer0 {
	#[cfg_attr(not(release), inline(never))]
	fn update(&mut self, beater: &Beater) -> Vec<Box<Enemy>> {
		if !beater.onsubbeat() {
			return vec![];
		}
		self.subbeats += 1;
		match self.subbeats % 16 { // TODO
			0 => {
				self.y = (rand() % 400) as i16 - 200;
				vec![]
			}
			4 => vec![ Box::new(EnemyE::new(&self.models, (300, self.y, -0x1000))), ],
			5 => vec![ Box::new(EnemyE::new(&self.models, (500, self.y, -0x1000))), ],
			6 => vec![ Box::new(EnemyE::new(&self.models, (700, self.y, -0x1000))), ],
			7 => vec![ Box::new(EnemyE::new(&self.models, (900, self.y, -0x1000))), ],
			12 => vec![ Box::new(EnemyE::new(&self.models, (-300, self.y, -0x1000))), ],
			13 => vec![ Box::new(EnemyE::new(&self.models, (-500, self.y, -0x1000))), ],
			14 => vec![ Box::new(EnemyE::new(&self.models, (-700, self.y, -0x1000))), ],
			15 => vec![ Box::new(EnemyE::new(&self.models, (-900, self.y, -0x1000))), ],
			_ => vec![]
		}
	}
}

pub struct Lockon {
	enemy: Weak<RefCell<Box<Enemy>>>, // TODO should be Targetable (=Enemy+Schematics+Overdrive)
	locked: bool,
	lockframe: i32,
	lockeffframe: i32,
	lockcount: u32,
	committed: bool,
	enforcing: bool,
	enforceframe: i32,
	enforcecount: u32,
}

impl Lockon {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(enemy: &Rc<RefCell<Box<Enemy>>>) -> Lockon {
		Lockon {
			enemy: Rc::downgrade(enemy),
			locked: false,
			lockframe: 0,
			lockeffframe: 0,
			lockcount: 0,
			committed: false,
			enforcing: false,
			enforceframe: 0,
			enforcecount: 0,
		}
	}
	#[cfg_attr(not(release), inline(never))]
	pub fn is_valid(&self) -> bool {
		self.enemy.upgrade().map(|re|re.borrow().alive()).unwrap_or(false) && // enemy must alive, and
			!(self.locked && self.lockcount <= self.enforcecount) // not enforced all
	}
	#[cfg_attr(not(release), inline(never))]
	pub fn lock(&mut self) -> bool {
		if !self.is_valid() { return false; }
		let rc_enemy = self.enemy.upgrade().unwrap();
		if !rc_enemy.borrow().can_lockon() { return false; }
		if self.lockcount != 0 && self.lockeffframe < 4 { return false; } // too quick additional lock is ignored
		self.lockcount += 1;
		self.lockeffframe = 0;
		rc_enemy.borrow_mut().lockon();
		return true;
	}
	#[cfg_attr(not(release), inline(never))]
	pub fn draw(&self) {
		let frame_time = 15;
		let frame_add = 3;
		let enforce_time = 15;

		// FIXME upgrade/unwrap enemy twice...
		// TODO this should be removed... it requires two Lockon GCs?
		if !self.is_valid() { return; }

		let rc_enemy = self.enemy.upgrade().unwrap(); // FIXME for lifetime issue!
		let cpos = rc_enemy.borrow().cached_spos();

		// frame
		if self.lockeffframe < frame_time {
			let draw_frame = |fr: i32| {
				let (sx, sy) = ((frame_time - fr) * 130 / frame_time + 20, (frame_time - fr) * 80 / frame_time + 20);
				let t = Qang::from_raw((-fr + frame_add) * Qang::one() / 8 / (frame_time - frame_add));
				let d135 = Qang::from_deg(135);
				let d45 = Qang::from_deg(45);
				gpu_opq_shaded_polyline(&[
					((cpos.0 + cos(t + d135).muli32(sx) as i16, cpos.1 - sin(t + d135).muli32(sy) as i16), (0, 0, 0)),
					((cpos.0 + cos(t + d45).muli32(sx) as i16, cpos.1 - sin(t + d45).muli32(sy) as i16), (255, 255, 255)),
					((cpos.0 + cos(t - d45).muli32(sx) as i16, cpos.1 - sin(t - d45).muli32(sy) as i16), (0, 0, 0)),
					((cpos.0 + cos(t - d135).muli32(sx) as i16, cpos.1 - sin(t - d135).muli32(sy) as i16), (255, 255, 255)),
					((cpos.0 + cos(t + d135).muli32(sx) as i16, cpos.1 - sin(t + d135).muli32(sy) as i16), (0, 0, 0)),
					]);
			};
			draw_frame(self.lockeffframe);
			if 1 <= self.lockeffframe {
				draw_frame(self.lockeffframe - 1);
			}
		}
		// icon
		if self.locked {
			// note: texpage must be done as SemiTrans::Mix, page=(9, 0)
			gpu_semitrans_mono_rect((0, 0, 0), (cpos.0 - 11, cpos.1 - 9), (24, 19));
			gpu_opq_raw_rect((cpos.0 - 11, cpos.1 - 9), (24, 19), (0, if self.committed { 20 } else { 0 }), (0, 0));
		}
		// humanoid to enemy line
		{
			let (start_num, start_denom, end_num, end_denom) =
				if !self.locked {
					(0, 1, self.lockframe, frame_time)
				} else if self.enforcing && self.enforcecount + 1 == self.lockcount { // last enforcing
					(self.enforceframe, enforce_time, 1, 1)
				} else {
					(0, 1, 1, 1)
				};
			let enemy_rc = self.enemy.upgrade().unwrap(); // just for lifetime
			let enepos = enemy_rc.borrow().pos();
			let mypos = (0, -200, 0);
			// FIXME ugly vector manipulations...
			let posdiff = ((enepos.0 - mypos.0) as i32, (enepos.1 - mypos.1) as i32, (enepos.2 - mypos.2) as i32);
			let startpos = (
				(mypos.0 as i32 + posdiff.0 * start_num / start_denom) as i16,
				(mypos.1 as i32 + posdiff.1 * start_num / start_denom) as i16,
				(mypos.2 as i32 + posdiff.2 * start_num / start_denom) as i16);
			let endpos = (
				(mypos.0 as i32 + posdiff.0 * end_num / end_denom) as i16,
				(mypos.1 as i32 + posdiff.1 * end_num / end_denom) as i16,
				(mypos.2 as i32 + posdiff.2 * end_num / end_denom) as i16);
			// yes, translate... enemy spos cache is meaningless.
			let (startsxy, _) = xlate_to_screen(startpos);
			let (endsxy, _) = xlate_to_screen(endpos);
			gpu_opq_mono_line((255, 128, 0), ((startsxy & 0xFFFF) as i16, (startsxy >> 16) as i16), ((endsxy & 0xFFFF) as i16, (endsxy >> 16) as i16));
		}
	}
	// TODO returning hit looks bad spec...
	#[cfg_attr(not(release), inline(never))]
	pub fn update(&mut self) -> bool {
		let frame_time = 15;
		let enforce_time = 15;

		if !self.locked {
			self.lockframe += 1;
			if frame_time <= self.lockframe {
				self.locked = true;
			}
		}
		self.lockeffframe += 1; // TODO beware of overflowing
		if self.enforcing {
			self.enforceframe += 1;
			if enforce_time <= self.enforceframe {
				//self.enemy.upgrade().unwrap().borrow_mut().hit(); // this and sound is others (scene/musicseq) work.
				self.enforcing = false;
				self.enforcecount += 1;
				return true;
			}
		}
		return false;
	}
	// if already committed, no effect.
	#[cfg_attr(not(release), inline(never))]
	pub fn commit(&mut self) {
		self.committed = true;
	}
	#[cfg_attr(not(release), inline(never))]
	pub fn enforce(&mut self) {
		// TODO activate. this is called from Scene when next beat is attack timing.
		self.enforcing = true;
		self.enforceframe = 0;
	}

	pub fn enemy(&self) -> Weak<RefCell<Box<Enemy>>> { self.enemy.clone() }
	pub fn committed(&self) -> bool { self.committed }
	pub fn enforcing(&self) -> bool { self.enforcing }
}

pub struct ShootingScene {
	frame: u32,
	beater: Beater,
	models: Rc<Models>, // TODO ugh, or reference hell with EnemySequencer?
	sounds: Rc<Sounds>,
	musicseq: Box<MusicSequencer>,
	enemyseq: Box<EnemySequencer>,
	btns: DigitalPadButtonState,
	peaker: Peaker,
	humanscale: i32,
	sight: (i16, i16),
	sightcount: u32, // for speed. if not, lockon.count(|e| !e.committed() )
	sightrot: i32,
	camsight: (i32, i32), // q16
	tocamsight: (i32, i32), // q16
	enemies: Vec<Rc<RefCell<Box<Enemy>>>>, // FIXME I don't want to use RefCell... but Rc refuses !Sized get_mut!
	bullets: Vec<Box<Bullet>>,
	effects: Vec<Box<Effect>>,
	lockon: Vec<Lockon>,
	last_enforced_rhythm: u32,
	//tri: usize,
}

impl ShootingScene {
	#[cfg_attr(not(release), inline(never))]
	pub fn new() -> ShootingScene {
		let mut progress = Progress::new(14-7);
		let mut spu = SPUMemAllocator::new();

		let chat_addr = spu.load_file(b"cdrom:\\CLOSEHAT.VB;1\0");
		progress.step();
		//let bdrum_addr = spu.load_file(b"cdrom:\\BASSDRUM.VB;1\0");
		//progress.step();
		//let choir_addr = spu.load_file(b"cdrom:\\CHOIR.VB;1\0");
		//progress.step();
		let perc1_addr = spu.load_file(b"cdrom:\\PERC1.VB;1\0");
		progress.step();
		let snare_addr = spu.load_file(b"cdrom:\\ACCSNARE.VB;1\0");
		progress.step();
		//let hand_addr = spu.load_file(b"cdrom:\\HANDCRAP.VB;1\0");
		//progress.step();

		gpu_byteload(576, 0, 256, 64, binread(b"cdrom:\\MARK_15.RAW;1\0").expect("Can't load MARK_15.RAW"));
		progress.step();

		let humanoid = Rc::new(TDModel::load(b"cdrom:\\HUMANOID.BIN;1\0"));
		progress.step();
		//let enemya = Rc::new(TDModel::load(b"cdrom:\\ENEMYA.BIN;1\0"));
		//progress.step();
		//let enemyc = Rc::new(TDModel::load(b"cdrom:\\ENEMYC.BIN;1\0"));
		//progress.step();
		//let enemyd = Rc::new(TDModel::load(b"cdrom:\\ENEMYD.BIN;1\0"));
		//progress.step();
		let enemye = Rc::new(TDModel::load(b"cdrom:\\ENEMYE.BIN;1\0"));
		progress.step();
		//let bulleta = Rc::new(TDModel::load(b"cdrom:\\BULLETA.BIN;1\0"));
		//progress.step();
		//let bulletb = Rc::new(TDModel::load(b"cdrom:\\BULLETB.BIN;1\0"));
		//progress.step();

		// Space Echo
		unsafe {
			let params: [u16; 32] = [
				0x033D,0x0231,0x7E00,0x5000,0xB400,0xB000,0x4C00,0xB000,
				0x6000,0x5400,0x1ED6,0x1A31,0x1D14,0x183B,0x1BC2,0x16B2,
				0x1A32,0x15EF,0x15EE,0x1055,0x1334,0x0F2D,0x11F6,0x0C5D,
				0x1056,0x0AE1,0x0AE0,0x07A2,0x0464,0x0232,0x8000,0x8000
			];
			::std::intrinsics::volatile_store(0x1F801DA2 as *mut u16, 0xE128);
			for (i,p) in params.iter().enumerate() {
				::std::intrinsics::volatile_store((0x1F801DC0 + i * 2) as *mut u16, *p);
			}
		}
		spucnt_unmute(true);
		spucnt_reverb(true);
		spu_vol((0x3FFF, 0x3FFF), (0x5EBC, 0x5EBC));

		// initialize effect area with bit15=0
		// note: fillvram/nontexpoly|rect(really?) fills with bit15=0, so can't use it.
		//       instead, use space area of loaded texture...
		//       to enlarge small area, can't use texrect... use texquad.
		// FIXME texwindow and rect should be faster? but this is one-time, it is ok...
		{
			set_draw_area((320, 0), (320-1, 240*2-1));
			gp0(0xE600_0001); // force_bit15=1, drawing_mask=0(always_draw)
			gpu_opq_raw15_quad((0, 0), (320, 0), (0, 240*2), (320, 240*2), (9, 0), (62, 0), (62, 0), (62, 0), (62, 0));
		}

		let models = Rc::new(Models {
				humanoid: humanoid,
				//enemya: enemya,
				//enemyc: enemyc,
				//enemyd: enemyd,
				enemye: enemye,
				//bulleta: bulleta,
				//bulletb: bulletb,
			});

		let sounds = Rc::new(Sounds {
				closehat: chat_addr,
				//bassdrum: bdrum_addr,
				//choir: choir_addr,
				perc1: perc1_addr,
				accsnare: snare_addr,
				//handcrap: hand_addr,
			});

		let mut beater = Beater::new(85);
		let musicseq = Box::new(MusicSequencerLayer0::new(sounds.clone(), &mut beater));
		let first_rhythm = musicseq.current_rhythm();

		ShootingScene {
			frame: 0,
			beater: beater,
			models: models.clone(),
			sounds: sounds.clone(),
			musicseq: musicseq,
			enemyseq: Box::new(EnemySequencerLayer0::new(models.clone())),
			btns: Default::default(),
			peaker: Peaker::new(5, 10, 1, 10, 1),
			humanscale: 0x0500,
			sight: (160, 120),
			sightcount: 0,
			sightrot: 0,
			camsight: (0, 0),
			tocamsight: (0, 0),
			enemies: vec![],
			bullets: vec![],
			effects: vec![],
			lockon: vec![],
			last_enforced_rhythm: first_rhythm,
			//tri: 0,
		}
	}

}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn mult_load_mat9_rtll(m: &[i16; 9]) {
	mult_load_mat9_rt(m);
	mult_load_mat9_ll(m);
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn mat_xrotate(theta: Qang) -> [i16; 9] {
	let c = cos(theta).raw_i16();
	let s = sin(theta).raw_i16();
	return [ 0x1000, 0, 0, 0, c, -s, 0, s, c ];
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn mat_yrotate(theta: Qang) -> [i16; 9] {
	let c = cos(theta).raw_i16();
	let s = sin(theta).raw_i16();
	return [ c, 0, s, 0, 0x1000, 0, -s, 0, c ];
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn mat_zrotate(theta: Qang) -> [i16; 9] {
	let c = cos(theta).raw_i16();
	let s = sin(theta).raw_i16();
	return [ c, s, 0, -s, c, 0, 0, 0, 0x1000 ];
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn mat_scale(sx: i16, sy: i16, sz: i16) -> [i16; 9] {
	return [ sx, 0, 0, 0, sy, 0, 0, 0, sz ];
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn mat_scale1(sf: i16) -> [i16; 9] {
	return [ sf, 0, 0, 0, sf, 0, 0, 0, sf ];
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn rotscale_by_lc(tx: i16, ty: i16, tz: i16) -> (i32, i32, i32) {
	// load to ir
	ir::set(Vec3Q12 { x: tx, y: ty, z: tz });
	// xlate by LC
	mvmva!(1, 2, 3, 3, 0); // sf=1 LC*IR+0(->MACv,IRv) signed-clamp
	return (mfc2!(25), mfc2!(26), mfc2!(27));
}

#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn xlate_by_rt(tx: i16, ty: i16, tz: i16) -> (i32, i32, i32) {
	// load to ir
	ir::set(Vec3Q12 { x: tx, y: ty, z: tz });
	// xlate by RT+TR
	mvmva!(1, 0, 3, 0, 0); // sf=1 RT*IR+TR(->MACv,IRv) signed-clamp
	return (mfc2!(25), mfc2!(26), mfc2!(27));
}

// can translate i16 range at once. (not i32...)
#[allow(dead_code)]
#[cfg_attr(not(release), inline(never))]
fn translate_rt((tx, ty, tz): (i16, i16, i16)) {
	// copy MACv to TR
	// bouncing to avoid "unnecessary unsafe block" warning...
	let (vx, vy, vz) = xlate_by_rt(tx, ty, tz);
	mtc2!(37, vx);
	mtc2!(38, vy);
	mtc2!(39, vz);
}

impl Scene for ShootingScene {
	#[cfg_attr(not(release), inline(never))]
	fn draw(&self) {
		// VRAM layout:
		// + 0,0---+ 320,0 --+ 576,0
		// | main1 | eff1    |
		// +-0,240-+ ...     +
		// + 0,256-+ 320,256-+
		// | main2 + eff2    |
		// +-------+ ...     + 576,512
		//  0,496  +---------+

		// switch displaying frame: this must be first.
		{
			let preyoff = self.frame * 256;
			gp1(0x05000000 | (  0 << 0) | (preyoff << 10));
		}

		// setup
		//gp0(0xE1000000 | gpu_calc_texpage((9, 0), SemiTrans::Mix, 2)); // texpage (for rects and mono-tris/quads)
		gp0(0xE2000000); // texwindow
		//gp0(0xE6000000); // maskbits

		// prepare OT and GTE regs...
		let mut ot = OrderTab::new(1024, 500/*just guess*/);

		// setup GTE regs...
		mtc2!(45, 0x00000100); mtc2!(46, 0x00000100); mtc2!(47, 0x00000200); // BKq12?
		mtc2!(53, 0x00000000); mtc2!(54, 0x00000000); mtc2!(55, 0x00000000); // FCq4
		// for RTP
		mtc2!(56, 160 << 16); mtc2!(57, 120 << 16); // OFX,OFYq16
		mtc2!(58, 0x0400); // Hq0u16
		mtc2!(59, 0xFC82); // DQAq8 (coeff)
		mtc2!(60, 0x01400000); // DQBq24? (offset)
		// for AVSZ
		mtc2!(61, 0x0155>>1); // ZSF3q12 (1/3 / 2)
		mtc2!(62, 0x0100>>1); // ZSF4q12 (1/4 / 2)

		// initial
		load_mat9_rt(&[0x1000, 0, 0, 0, 0x1000, 0, 0, 0, 0x1000]);
		load_mat9_ll(&[0x1000, 0, 0, 0, 0x1000, 0, 0, 0, 0x1000]);
		mtc2!(37, 0); mtc2!(38, 0); mtc2!(39, 0); // TR = 0

		// scale z for perspective... TODO
		//mult_load_mat9_rt(&[ 0x1000, 0, 0, 0, 0x1000, 0, 0, 0, 0x1234, ]);

		// projectionview
		//mult_load_mat9_rt(&mat_xrotate(Qang::from_deg(10)));
		// look through slightly above humanoid
		{
			mult_load_mat9_rt(&mat_xrotate(Qang::from_deg(180))); // OpenGL->PS
			// cheat: LC as temporary matrix
			load_mat9_lc(&[0x1000, 0, 0, 0, 0x1000, 0, 0, 0, 0x1000]);
			mult_load_mat9_lc(&mat_yrotate(Qang::from_raw(((-self.camsight.0 >> 8) * Qang::one() / 180 / 80) >> 8)));
			mult_load_mat9_lc(&mat_xrotate(Qang::from_raw(((-self.camsight.1 >> 8) * Qang::one() / 180 / 80) >> 8)));
			let (x,y,z) = rotscale_by_lc(0, 0, 0x1000);
			let (m,t) = cpu_lookat_mat9tr(
				Vec3wQ12{x: x, y: y, z: z},
				Vec3wQ12{x: 0, y: 0, z: 0},
				Vec3wQ12{x: 0, y: 100, z: 0});
			translate_rt((t[0] as i16, t[1] as i16, t[2] as i16)); mult_load_mat9_rt(&m);
		}

		// light direction: initially lights Z- (far to near) that is defined by LC.
		//mult_load_mat9_ll(&mat_xrotate(Qang::from_deg(180))); // if no lookat(OGL->PS), use this.

		// setup LC after using "gte_normalize_ir_lc" (that brokes LC) before "enqueue"
		// color(1.0, 1.0, 1.0) only for (LL-xlated) Z+
		load_mat9_lc(&[0, 0, 0x1000, 0, 0, 0x1000, 0, 0, 0x1000]);

		// effect TODO impl later!
		// note: projection view must be set before this, because some effect depends to it.
		{
			let import_screen = false;

			gp0(0xE6000001); // maskbits: bit15=always1 (important for both copy-imported-screen and copy-prev-effect)

			// import current displaying screen (prev.draw) to tmp area
			// 320,240 --1/2-> 160(128+32),120 divided because texture is max 256x256: 320=256+64
			if import_screen {
				set_draw_area((576, 256), (576 + 160-1, 256 + 120-1));

				// left-almost
				gpu_opq_raw15_quad(
					(0, 0), (128, 0), (0, 120), (128, 120),
					(0, self.frame), // cur.disp left (prev.draw)
					(0, 0), (255, 0), (0, 240), (255, 240));
				// right-slightly
				gpu_opq_raw15_quad(
					(128, 0), (160, 0), (128, 120), (160, 120),
					(4, self.frame), // cur.disp right (prev.draw)
					(0, 0), (64, 0), (0, 240), (64, 240));
			}

			// set drawarea to drawing effect area
			{
				let curyoff = (1 - self.frame) * 256;
				set_draw_area((320, curyoff), (320 + 256-1, curyoff + 256-1));
			}

			// copy imported-screen to center of effect area?
			if import_screen {
				//gp0(0xE1000000 | gpu_calc_texpage((9, 1), SemiTrans::Add, 2)); // texpage=9,1(imported-screen), SemiTrans::Add.
				//gpu_semitrans_texblend_rect((0x30, 0x30, 0x30), (48, 68), (160, 120), (0, 0), (0, 0));
				gp0(0xE1000000 | gpu_calc_texpage((9, 1), SemiTrans::Mix, 2)); // texpage=9,1(imported-screen)
				gpu_opq_raw_rect((48, 68), (160, 120), (0, 0), (0, 0));
			}

			// copy previous with enlargement and dark
			{
				let left = -8;
				let top = -8;
				let right = 256+8;
				let bottom = 256+8;
				gpu_semitrans_texblend15_quad((0x70, 0x70, 0x70),
					(left, top), (right, top), (left, bottom), (right, bottom),
					(5, self.frame), // previous effect
					(0, 0), (255, 0), (0, 255), (255, 255),
					SemiTrans::Mix);
			}

			// draw effects (effect_area)
			{
				let mat = get_mat33trgte_rt();
				for e in &self.effects {
					translate_rt(e.pos());
					e.draw_effectarea();
					load_mat33trgte_rt(&mat);
				}
			}
		}
		gp0(0xE1000000 | gpu_calc_texpage((9, 0), SemiTrans::Mix, 2)); // reset texpage
		gp0(0xE6000000); // reset maskbits

		// switch drawing area to displaying area
		{
			let curyoff = (1 - self.frame) * 256;
			set_draw_area((0, curyoff), (320-1, curyoff + 240-1));
			// fillvram is not affected by drawoffset...
			fillvram(16, 16, 32, 0, curyoff as u16, 320, 240);
		}

		// import effect
		{
			gpu_opq_raw15_quad(
				(0, 0), (320, 0), (0, 240), (320, 240),
				(5, 1 - self.frame),
				(48, 68), (48+160, 68), (48, 68+120), (48+160, 68+120)
				);
		}
		// draw background
		{
			let mat = get_mat33trgte_rt();
			translate_rt((0, -200, 0));
			self.peaker.draw();
			load_mat33trgte_rt(&mat);
		}
		// draw effects (before)
		{
			let mat = get_mat33trgte_rt();
			for e in &self.effects {
				translate_rt(e.pos());
				e.draw_before();
				load_mat33trgte_rt(&mat);
			}
		}

		// settle OT before inserting
		ot.settle();

		// place models
		{
			let mat = get_mat33trgte_rt();
			translate_rt((0, -200, 0));
			mult_load_mat9_rt(&mat_scale1(self.humanscale as i16));
			mult_load_mat9_rtll(&mat_xrotate(Qang::from_deg(30)));
			ot.enqueue(&self.models.humanoid);
			load_mat33trgte_rt(&mat);

			for re in &self.enemies {
				{
					let e = re.borrow();
					translate_rt(e.pos());
					e.draw(&mut ot);

					// FIXME mutating immut enemy!
					let (sxy, _) = xlate_to_screen(e.pos());
					e.update_spos_cache(((sxy & 0xFFFF) as i16, (sxy >> 16) as i16));
				}
				load_mat33trgte_rt(&mat);
			}

			// draw effects (inter)
			{
				let mat = get_mat33trgte_rt();
				for e in &self.effects {
					translate_rt(e.pos());
					e.draw_inter(&mut ot);
					load_mat33trgte_rt(&mat);
				}
			}
		}

		// draw!
		ot.submit();
		ot.wait(); // required on XEBRA 150831/ePSXe 2.0.5 if poking GP0 after this.

		// draw post-3D

		// draw effects (after)
		{
			let mat = get_mat33trgte_rt();
			for e in &self.effects {
				translate_rt(e.pos());
				e.draw_after();
				load_mat33trgte_rt(&mat);
			}
		}

		// draw lockon effects
		gp0(0xE1000000 | gpu_calc_texpage((9, 0), SemiTrans::Mix, 2));
		for e in &self.lockon { e.draw(); }

		// draw sight
		// TODO drawoffset is usable here?
		{
			let points = [
									 (self.sight.0 - 20, self.sight.1 - 15), // topleft
									 (self.sight.0 - 15, self.sight.1 - 20), // ->rightup
									 (self.sight.0 + 15, self.sight.1 - 20), // ->topright
									 (self.sight.0 + 20, self.sight.1 - 15), // ->rightdown
									 (self.sight.0 + 20, self.sight.1 + 15), // ->bottomright
									 (self.sight.0 + 15, self.sight.1 + 20), // ->leftdown
									 (self.sight.0 - 15, self.sight.1 + 20), // ->bottomleft
									 (self.sight.0 - 20, self.sight.1 + 15), // ->leftup
									 (self.sight.0 - 20, self.sight.1 - 15), // ->topleft
				];
			// darkarea
			gp0(0xE1000000 | gpu_calc_texpage((9, 0), SemiTrans::Mix, 2));
			for i in 0..(points.len() - 1) {
				// 22: mono-tri-semitrans
				gpu_monotri(0x22, (0, 0, 0), points[i], points[i + 1], self.sight);
			}
			// frame
			gpu_opq_mono_polyline((255, 255, 255), &points);
			// cross
			{
				let t = Qang::from_raw(self.sightrot);
				let c = cos(t).muli32(3) as i16;
				let s = sin(t).muli32(3) as i16;
				gpu_opq_mono_line((255, 255, 255), (self.sight.0 + c, self.sight.1 - s), (self.sight.0 - c, self.sight.1 + s));
			}
			{
				let t = Qang::from_raw(self.sightrot + Qang::one() / 4);
				let c = cos(t).muli32(3) as i16;
				let s = sin(t).muli32(3) as i16;
				gpu_opq_mono_line((255, 255, 255), (self.sight.0 + c, self.sight.1 - s), (self.sight.0 - c, self.sight.1 + s));
			}
			// count: ~18x20
			// TODO should texture all? how line drawn is different by emulators... how line vs texrect speed?
			match self.sightcount {
				1 => {
					/*
					// to make life easier when drawoffset is used, I want to divide ideal-pos and offsetting-by-self.sight
					// but Rust do not allow this...
					let mut tmp; // make more life by prepare this out of closure.
					gpu_opq_mono_polyline((255, 255, 255), [
						(0, 0),
					].iter().map(|&(x, y): &(i16, i16)| { tmp = (self.sight.0 + x, self.sight.1 + y); &tmp })); // lifetime what?
					*/
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 6, self.sight.1 - 11),
						(self.sight.0 + 2, self.sight.1 - 11),
						(self.sight.0 + 2, self.sight.1 + 8),
						(self.sight.0 + 6, self.sight.1 + 8),
						(self.sight.0 + 6, self.sight.1 + 11),
						(self.sight.0 - 6, self.sight.1 + 11),
						(self.sight.0 - 6, self.sight.1 + 8),
						(self.sight.0 - 1, self.sight.1 + 8),
						(self.sight.0 - 1, self.sight.1 - 8),
						(self.sight.0 - 6, self.sight.1 - 8),
						(self.sight.0 - 6, self.sight.1 - 11),
					]);
				},
				2 => {
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 +  1),
						(self.sight.0 - 6, self.sight.1 +  1),
						(self.sight.0 - 6, self.sight.1 +  7),
						(self.sight.0 + 9, self.sight.1 +  7),
						(self.sight.0 + 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 -  2),
						(self.sight.0 + 6, self.sight.1 -  2),
						(self.sight.0 + 6, self.sight.1 -  7),
						(self.sight.0 - 9, self.sight.1 -  7),
						(self.sight.0 - 9, self.sight.1 - 10),
					]);
				},
				3 => {
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 +  7),
						(self.sight.0 + 6, self.sight.1 +  7),
						(self.sight.0 + 6, self.sight.1 +  1),
						(self.sight.0 - 9, self.sight.1 +  1),
						(self.sight.0 - 9, self.sight.1 -  2),
						(self.sight.0 + 6, self.sight.1 -  2),
						(self.sight.0 + 6, self.sight.1 -  7),
						(self.sight.0 - 9, self.sight.1 -  7),
						(self.sight.0 - 9, self.sight.1 - 10),
					]);
				},
				4 => {
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 9, self.sight.1 - 10),
						(self.sight.0 - 6, self.sight.1 - 10),
						(self.sight.0 - 6, self.sight.1 -  1),
						(self.sight.0 + 6, self.sight.1 -  1),
						(self.sight.0 + 6, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 + 10),
						(self.sight.0 + 6, self.sight.1 + 10),
						(self.sight.0 + 6, self.sight.1 +  2),
						(self.sight.0 - 9, self.sight.1 +  2),
						(self.sight.0 - 9, self.sight.1 - 10),
					]);
				},
				5 => {
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 -  7),
						(self.sight.0 - 6, self.sight.1 -  7),
						(self.sight.0 - 6, self.sight.1 -  2),
						(self.sight.0 + 9, self.sight.1 -  2),
						(self.sight.0 + 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 +  7),
						(self.sight.0 + 6, self.sight.1 +  7),
						(self.sight.0 + 6, self.sight.1 +  1),
						(self.sight.0 - 9, self.sight.1 +  1),
						(self.sight.0 - 9, self.sight.1 - 10),
					]);
				},
				6 => {
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 - 10),
						(self.sight.0 + 9, self.sight.1 -  7),
						(self.sight.0 - 6, self.sight.1 -  7),
						(self.sight.0 - 6, self.sight.1 -  2),
						(self.sight.0 + 9, self.sight.1 -  2),
						(self.sight.0 + 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 + 10),
						(self.sight.0 - 9, self.sight.1 - 10),
					]);
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 6, self.sight.1 +  1),
						(self.sight.0 + 6, self.sight.1 +  1),
						(self.sight.0 + 6, self.sight.1 +  7),
						(self.sight.0 - 6, self.sight.1 +  7),
						(self.sight.0 - 6, self.sight.1 +  1),
					]);
				},
				7 => {
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 7, self.sight.1 - 10),
						(self.sight.0 + 7, self.sight.1 - 10),
						(self.sight.0 + 7, self.sight.1 + 10),
						(self.sight.0 + 4, self.sight.1 + 10),
						(self.sight.0 + 4, self.sight.1 -  7),
						(self.sight.0 - 7, self.sight.1 -  7),
						(self.sight.0 - 7, self.sight.1 - 10),
					]);
				},
				8 => { // "max"
					// cheat: texture
					gpu_opq_raw_rect((self.sight.0 - 18, self.sight.1 - 7), (37, 15), (25, 0), (0, 0));
					/*
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 - 18, self.sight.1 -  6),
						(self.sight.0 -  6, self.sight.1 -  6),
						(self.sight.0 -  6, self.sight.1 +  7),
						(self.sight.0 -  8, self.sight.1 +  7),
						(self.sight.0 -  8, self.sight.1 -  4),
						(self.sight.0 - 11, self.sight.1 -  4),
						(self.sight.0 - 11, self.sight.1 +  7),
						(self.sight.0 - 13, self.sight.1 +  7),
						(self.sight.0 - 13, self.sight.1 -  4),
						(self.sight.0 - 16, self.sight.1 -  4),
						(self.sight.0 - 16, self.sight.1 +  7),
						(self.sight.0 - 18, self.sight.1 +  7),
						(self.sight.0 - 18, self.sight.1 -  6),
					]);
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 -  4, self.sight.1 -  6),
						(self.sight.0 +  5, self.sight.1 -  6),
						(self.sight.0 +  5, self.sight.1 +  7),
						(self.sight.0 -  4, self.sight.1 +  7),
						(self.sight.0 -  4, self.sight.1 +  0),
						(self.sight.0 +  2, self.sight.1 +  0),
						(self.sight.0 +  2, self.sight.1 -  4),
						(self.sight.0 -  2, self.sight.1 -  4),
						(self.sight.0 -  2, self.sight.1 -  2),
						(self.sight.0 -  4, self.sight.1 -  2),
						(self.sight.0 -  4, self.sight.1 -  6),
					]);
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 -  2, self.sight.1 +  2),
						(self.sight.0 +  2, self.sight.1 +  2),
						(self.sight.0 +  2, self.sight.1 +  5),
						(self.sight.0 -  2, self.sight.1 +  5),
						(self.sight.0 -  2, self.sight.1 +  2),
					]);
					gpu_opq_mono_polyline((255, 255, 255), &[
						(self.sight.0 +  7, self.sight.1 -  6),
						(self.sight.0 +  9, self.sight.1 -  6),
						(self.sight.0 + 13, self.sight.1 -  1), // 13,-1? (x pair with 3.x)
						(self.sight.0 + 15, self.sight.1 -  6),
						(self.sight.0 + 17, self.sight.1 -  6),
						(self.sight.0 + 17, self.sight.1 -  4),
						(self.sight.0 + 13, self.sight.1 +  1), // 13,1? (y pair with 4.y)
						(self.sight.0 + 16, self.sight.1 +  5),
						(self.sight.0 + 16, self.sight.1 +  7),
						(self.sight.0 + 14, self.sight.1 +  7),
						(self.sight.0 + 13, self.sight.1 +  2), // 13,2?
						(self.sight.0 +  9, self.sight.1 +  7),
						(self.sight.0 +  7, self.sight.1 +  7),
						(self.sight.0 +  7, self.sight.1 +  5),
						(self.sight.0 +  9, self.sight.1 +  1), //  9,1?
						(self.sight.0 +  7, self.sight.1 -  4),
						(self.sight.0 +  7, self.sight.1 -  6),
					]);
					*/
				},
				_ => ()/*nothing*/
			};
		}

		// draw HUD TODO
	}
	#[cfg_attr(not(release), inline(never))]
	fn input(&mut self, bits: u16) -> Option<Box<Scene>> {
		self.btns.update(bits);
		let press = self.btns.press();

		let speed = 70 * self.beater.bpm() / 60;

		// beat core
		self.beater.update();
		// then beat sequencer and backgrounds
		self.musicseq.update(&mut self.beater);
		self.enemies.extend(self.enemyseq.update(&self.beater).into_iter().map(|e| Rc::new(RefCell::new(e)) ));
		self.peaker.update(&self.beater, speed);

		for e in &mut self.enemies { self.bullets.extend(e.borrow_mut().update(&self.beater)) }
		self.lockon.retain(|o| o.is_valid() ); // must be done before killing enemy entry TODO ugly!!!
		self.enemies.retain(|re| re.borrow().alive() );

		for e in &mut self.bullets { e.update() }

		for e in &mut self.effects { e.update(); }
		self.effects.retain(|e| e.alive() );

		// enforce
		for e in &mut self.lockon {
			if e.committed() && !e.enforcing() {
				let next_rid = self.musicseq.next_rhythm_in(&self.beater, max(self.musicseq.current_rhythm(), self.last_enforced_rhythm), 15);
				if let Some(rid) = next_rid {
					e.enforce();
					self.last_enforced_rhythm = rid;
					break; // at most one enforce is done in a frame. TODO some cases are not! pair in enemy group, not rhythm group?
				}
			}
		}

		// step lockon and check enemy alive if an enforcing is done
		for e in &mut self.lockon {
			if e.update() {
				// this lockon hits enemy. timing is synchronized by enforcing
				let enemy_rc = e.enemy().upgrade().unwrap(); // for lifetime
				let mut emut = enemy_rc.borrow_mut();
				emut.hit();
				spu_play_ex(1, self.sounds.accsnare, /*78564@C5->C3=/4=*/19641, (0x3FFF, 0x3FFF), 0x000F5F00, false);
				if !emut.alive() {
					// create broken effect
					self.effects.push(Box::new(SonicCircleEffect::new(emut.pos(), 350)));
					self.effects.push(Box::new(SoftBoomEffect::new(emut.pos())));
					// this lockon should be gc'd.
				}
			}
		}

		// move sight
		if press.up() { self.sight = (self.sight.0, max(40, self.sight.1 - 4)); }
		else if press.down() { self.sight = (self.sight.0, min(200, self.sight.1 + 4)); }
		if press.left() { self.sight = (max(40, self.sight.0 - 4), self.sight.1); }
		else if press.right() { self.sight = (min(280, self.sight.0 + 4), self.sight.1); }

		// enqueue lockon
		if press.cross() {
			for rre in &self.enemies {
				if 8 <= self.sightcount { break; }
				let (x, y) = rre.borrow().cached_spos();
				if self.sight.0 - 20 < x && x < self.sight.0 + 20 && self.sight.1 - 20 < y && y < self.sight.1 + 20 {
					/*
					fn ptr_eq<T>(a: &T, b: &T) -> bool { a as *const T == b as *const T }
					let lockon = match self.lockon.iter_mut().find(|e|e.enemy().upgrade().map(|re| ptr_eq(&re, &rre) ).unwrap_or(false)) {
						Some(e) => e,
						None => {
							let lockon = Lockon::new(&rre);
							// FIXME can't I merge push+get_that_reference?
							self.lockon.push(lockon);
							self.lockon.last_mut().unwrap() // WTF! iter_mut and push/last_mut clash about mut ref! even already known as None!
						}
					};
					*/
					let lockon = {
						fn ptr_eq<T, R: AsRef<T>>(a: &R, b: &R) -> bool { a.as_ref() as *const _ == b.as_ref() as *const _ }
						let i = (0..self.lockon.len()).find(|i| self.lockon[*i].enemy().upgrade().map(|lre| ptr_eq(&lre, &rre) ).unwrap_or(false) );
						match i {
							Some(i) => &mut self.lockon[i],
							None => {
								let lockon = Lockon::new(&rre);
								// FIXME can't I merge push+get_that_reference?
								self.lockon.push(lockon);
								self.lockon.last_mut().unwrap()
							}
						}
					};
					if lockon.lock() {
						self.sightcount += 1;
						self.musicseq.lockedon();
						// TODO sound if not too quick to beat
					}
				}
			}
		}

		// commit fire
		if self.btns.up().cross() {
			for e in &mut self.lockon {
				e.commit();
			}
			self.sightcount = 0;
		}

		/*
		if self.btns.down().triangle() {
			if self.tri < 8 {
				let hzs = [6090, 8154, 10255, 12180, 16308, 20511, 24360, 32617]; // A#2~D#5
				spu_play_ex(16 + self.tri as u8 % 4 * 2, self.sounds.choir, hzs[self.tri], (0x3FFF, 0x3FFF), 0x000F5F00, false);
				spu_play_ex(17 + self.tri as u8 % 4 * 2, self.sounds.choir, hzs[self.tri] as u32 * 997 / 1000, (0x3FFF, 0x3FFF), 0x000F5F00, false);
				self.tri += 1;
			}
		}
		*/

		// calculate scale
		{
			let targetscale = 0x0300 + self.beater.beat_lerp(0x80);
			let diff = targetscale - self.humanscale;
			let maxdiff = 50;
			self.humanscale = self.humanscale + max(-maxdiff, min(maxdiff, diff));
		}

		self.sightrot = (self.sightrot + Qang::one() * 80/*bpm*/ / 4/*quadrant*/ / 60/*sec*/ / 60/*fps*/) % Qang::one();

		// camera works
		{
			let next = |s: i16, qc: i32| -> i32 {
				let margin = 40;
				let qm = margin << 16;
				let qs = (s as i32) << 16;
				if qs < qc - qm {
					qc + (qs - (qc - qm))
				} else if qc + qm < qs {
					qc + (qs - (qc + qm))
				} else {
					qc
				}
			};
			let x = next(self.sight.0 - 160, self.tocamsight.0);
			let y = next(self.sight.1 - 120, self.tocamsight.1);
			self.tocamsight = (x, y);
		}
		self.camsight = (
			(self.tocamsight.0 + self.camsight.0 * 20) / 21,
			(self.tocamsight.1 + self.camsight.1 * 20) / 21
			);

		self.frame = (self.frame + 1) % 2;

		return None;
	}
}
