extern crate libc;
use self::libc::{c_char, c_int, size_t, c_void};
use std;
use std::error::Error;
use std::fmt;
use std::io;
use std::io::Write;

extern "C" {
	pub fn calla0(nr: u8, ...) -> size_t;
	pub fn callb0(nr: u8, ...) -> size_t;
	pub fn callc0(nr: u8, ...) -> size_t;
	pub fn syscall(nr: u8);

	// no memcpy in liblibc, though there are memcmp and memchr.
	pub fn memcpy(dst: *mut c_void, src: *const c_void, size: size_t);
}

/// allocator: must treated by BIOS if use qsort in BIOS...
// note: XEBRA OSROM sometime returns non-word-aligned... I can't trust it!

#[cfg(use_bios_allocator)]
use self::biosallocator::*;
#[cfg(use_bios_allocator)]
pub mod biosallocator {
	use super::libc::{size_t, c_void};
	use super::{memcpy, calla0};

	#[inline(never)]
	#[no_mangle]
	pub extern "C" fn malloc(sz: size_t) -> *mut c_void {
		unsafe {
			calla0(0x33, sz) as *mut c_void
		}
	}

	#[inline(never)]
	#[no_mangle]
	pub extern fn realloc(addr: *mut c_void, size: size_t) -> *mut c_void {
		unsafe {
			// BIOS realloc has a bug, break allocated area!!
			//calla0(0x38, addr, size) as *mut c_void

			// instead, reuse if can, or malloc-memcpy-free.
			let chunk_size = *(addr as *mut usize).offset(-1);
			assert!(chunk_size & 1 == 0, "free-bit must be 0");
			if size <= chunk_size {
				// reuse.
				return addr;
			} else {
				// malloc, copy, free then return new addr.
				let new_addr = malloc(size);
				if new_addr.is_null() {
					return 0 as *mut c_void;
				}
				memcpy(new_addr, addr, chunk_size);
				free(addr);
				return new_addr;
			}
		}
	}

	#[inline(never)]
	#[no_mangle]
	pub extern fn free(addr: *mut c_void) {
		unsafe {
			calla0(0x34, addr);
		}
	}

	// posix_memalign is called if more than native align are required.
	// note: maybe called in panic.
	// MOVED TO start.s
	//#[inline(never)]
	//#[no_mangle]
	//#[allow(unused_variables)]
	//pub extern fn posix_memalign(addr: *mut c_void, align: size_t, size: size_t) -> ! {
	//	panic!("aligning more than malloc are not supported yet");
	//}

	// called from start.s
	#[inline(never)]
	#[no_mangle]
	pub extern fn init_heap(addr: *mut c_void, size: size_t) {
		unsafe {
			calla0(0x39, addr, size);
		}
	}
}

#[cfg(use_biosalign_allocator)]
use self::biosalignallocator::*;
#[cfg(use_biosalign_allocator)]
pub mod biosalignallocator {
	use super::libc::{size_t, c_void};
	use super::{memcpy, calla0};

	#[inline(never)]
	#[no_mangle]
	pub extern "C" fn malloc(sz: size_t) -> *mut c_void {
		let mut p = unsafe { calla0(0x33, sz + 8) as u32 };
		if p != 0 {
			let ap = p + 3 & !3;
			unsafe { *(ap as *mut u32) = p; }
			p = ap + 4;
		}
		return p as *mut c_void;
	}

	#[inline(never)]
	#[no_mangle]
	pub extern fn realloc(addr: *mut c_void, size: size_t) -> *mut c_void {
		unsafe {
			// BIOS realloc has a bug, break allocated area!!
			//calla0(0x38, addr, size) as *mut c_void

			// instead, reuse if can, or malloc-memcpy-free.
			let chunk_size = *(addr as *mut usize).offset(-1);
			assert!(chunk_size & 1 == 0, "free-bit must be 0");
			if size <= chunk_size {
				// reuse.
				return addr;
			} else {
				// malloc, copy, free then return new addr.
				let new_addr = malloc(size);
				if new_addr.is_null() {
					return 0 as *mut c_void;
				}
				memcpy(new_addr, addr, chunk_size);
				free(addr);
				return new_addr;
			}
		}
	}

	#[inline(never)]
	#[no_mangle]
	pub extern fn free(addr: *mut c_void) {
		let orgp = unsafe { *((addr as u32 - 4) as *const u32) };
		unsafe { calla0(0x34, orgp); }
	}

	// posix_memalign is called if more than native align are required.
	// note: maybe called in panic.
	// MOVED TO start.s
	//#[inline(never)]
	//#[no_mangle]
	//#[allow(unused_variables)]
	//pub extern fn posix_memalign(addr: *mut c_void, align: size_t, size: size_t) -> ! {
	//	panic!("aligning more than malloc are not supported yet");
	//}

	// called from start.s
	#[inline(never)]
	#[no_mangle]
	pub extern fn init_heap(addr: *mut c_void, size: size_t) {
		unsafe {
			calla0(0x39, addr, size);
		}
	}
}

// TODO so slow naive allocator... make some nice.
#[cfg(use_my_allocator)]
use self::myallocator::*;
#[cfg(use_my_allocator)]
pub mod myallocator {
	use super::libc::{size_t, c_void};
	use super::memcpy;

	static mut top: u32 = 0;

	#[cfg_attr(not(release), inline(never))]
	unsafe fn find_free(reqsz: u32) -> u32 { // sz must be roundup by 4
		let mut p = top;
		loop {
			let hdr = *(p as *const u32);
			if hdr == 0xFFFFffff {
				// reached end of heap; can't found any free block.
				return 0;
			}
			let size = hdr & !1; // size including header
			if hdr & 1 == 0 {
				// used block; go next
				p += size;
				continue;
			}
			if size - 4 < reqsz {
				// too small free block; go next
				p += size;
				continue;
			}
			// found enough free block; return it
			return p;
		}
	}

	#[cfg_attr(not(release), inline(never))]
	unsafe fn try_merge_free(p: u32) {
		let hdr = *(p as *const u32);
		if hdr & 1 == 0 {
			panic!("passed addr is not free block"); // FIXME cause double panic
		}
		let mut size = hdr & !1;
		loop {
			let nexthdr = *((p + size) as *const u32);
			let nextsize = nexthdr & !1;
			if nexthdr == 0xFFFFffff || nexthdr & 1 == 0 { // TODO "checking end" smells bugs...
				// next is used block; end.
				return;
			}
			// merge.
			size += nextsize;
			*(p as *mut u32) = size | 1;
			// try next.
		}
	}

	#[cfg_attr(not(release), inline(never))]
	unsafe fn try_split(p: u32, size: u32, reqsz: u32) { // size/reqsz must be roundup by 4
		if size < reqsz + 12 { // making free room for 4 bytes
			// small free; use whole block
			*(p as *mut u32) = size;
		} else {
			// enough large free; split
			*(p as *mut u32) = reqsz;
			*((p + reqsz) as *mut u32) = (size - reqsz) | 1; // rest is free
			// it is good time here to check merging next free, isn't it?
			try_merge_free(p + reqsz);
		}
	}

	#[inline(never)]
	#[no_mangle]
	pub extern "C" fn malloc(reqsz: size_t) -> *mut c_void {
		unsafe {
			let reqsz = 4 + ((reqsz as u32 + 3) & !3); // including header size, roundup
			let found = find_free(reqsz);
			if found == 0 {
				return 0 as *mut c_void;
			}
			// update; split or whole
			let header = *(found as *const u32);
			let size = header & !1; // size including header
			try_split(found, size, reqsz);
			return (found + 4) as *mut c_void;
		}
	}

	#[inline(never)]
	#[no_mangle]
	pub extern fn realloc(addr: *mut c_void, reqsz: size_t) -> *mut c_void {
		unsafe {
			let reqsz = 4 + ((reqsz as u32 + 3) & !3); // including header size, roundup
			let phdr = addr as u32 - 4; // I trust you, programmer!
			let header = *(phdr as *const u32);
			if header & 1 != 0 {
				// free!?
				panic!("passed addr is not valid memblock"); // FIXME cause double panic
			}
			let size = header & !1;
			if reqsz <= size {
				// making smaller or equal
				try_split(phdr, size, reqsz);
				return (phdr + 4) as *mut c_void;
			}
			// making bigger
			// opt: see just next block and try merge
			let nexthdr = *((phdr + *(phdr as *const u32)) as *const u32);
			let nextsize = nexthdr & !1;
			if nexthdr != 0xFFFFffff && nexthdr & 1 != 0 && (size + nextsize) <= reqsz + 12 { // TODO "checking end" smells bugs...
				// next is free block and have enough area; moving border
				try_split(phdr, size + nextsize, reqsz);
				return (phdr + 4) as *mut c_void;
			}
			// find bigger free block = malloc
			let found = malloc(reqsz as size_t);
			if found as u32 == 0 {
				// no enough free block found; die
				return 0 as *mut c_void;
			}
			// copy and free current
			memcpy(found, addr, (size - 4) as size_t);
			free(addr);
			return found;
		}
	}

	#[inline(never)]
	#[no_mangle]
	pub extern fn free(addr: *mut c_void) {
		unsafe {
			let addr = addr as u32 - 4; // I trust you, programmer!
			let p = addr as *mut u32;
			if *p & 1 != 0 {
				// already free!?
				panic!("passed addr is already freed memblock"); // FIXME cause double panic
			}
			*p |= 1;
			try_merge_free(addr);
		}
	}

	// posix_memalign is called if more than native align are required.
	// note: maybe called in panic.
	// MOVED TO start.s
	//#[inline(never)]
	//#[no_mangle]
	//#[allow(unused_variables)]
	//pub extern fn posix_memalign(addr: *mut c_void, align: size_t, size: size_t) -> ! {
	//	panic!("aligning more than malloc are not supported yet");
	//}

	// called from start.s
	#[inline(never)]
	#[no_mangle]
	pub extern fn init_heap(addr_p: *mut c_void, size_r: size_t) {
		unsafe {
			let addr = ((addr_p as u32) + 3) & !3; // roundup
			let size = (size_r as u32 - (addr - addr_p as u32)) & !3; // rounddown
			// initialize whole area as a free block
			*(addr as *mut u32) = (size - 4) | 1;
			*((addr + size - 4) as *mut u32) = 0xFFFFffff; // end of heap
			// remember top
			top = addr;
		}
	}
}

/// write and getenv are required by Rust runtime

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern fn write(_fd: c_int, src: *const c_void, size: size_t) {
	//printty(format!("stub: write: {} {}bytes to fd{}\n", src as size_t, size, fd).as_str());
	putstty(unsafe { std::slice::from_raw_parts(src as *const u8, size) });
}

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern fn getenv(pname: *const c_char) {
	let name: &str = unsafe { std::ffi::CStr::from_ptr(pname) }.to_str().unwrap();
	printty(format!("stub: getenv: {}\n", name).as_str());
}

/// pthread related functions that are required by Rust thread_local runtime (mainly panic)
/// note: assuming sizeof(pthread_key_t = c_int) == sizeof(void*), which is true for mipsel-unknown-linux-musl.
/// FIXME should maintain keys globally but pairs value thread-locally, but it isn't bothered for single-thread.

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern fn pthread_key_create(key: *mut libc::pthread_key_t, _dtor: Option<unsafe extern fn(*mut c_void)>) -> c_int {
	// TODO dtor?
	let p = malloc(std::mem::size_of::<*mut c_void>()) as libc::pthread_key_t;
	if p == 0 {
		return 1;
	}
	unsafe {
		*(p as *mut *const c_void) = std::ptr::null(); // initialize tls value with nullptr
		*key = p; // key is ptr to allocated *mut c_void !!ASSUMING pthread_key_t = c_int!!
	}
	return 0;
}

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern fn pthread_key_delete(key: libc::pthread_key_t) -> c_int {
	free(key as *mut c_void);
	return 0;
}

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern fn pthread_setspecific(key: libc::pthread_key_t, ptr: *const c_void) -> c_int {
	unsafe {
		*(key as *mut *const c_void) = ptr;
	}
	return 0;
}

#[cfg_attr(not(release), inline(never))]
#[no_mangle]
pub extern fn pthread_getspecific(key: libc::pthread_key_t) -> *mut c_void {
	unsafe {
		return *(key as *const *mut c_void);
	}
}

/// libc or BIOS

/*
#[no_mangle]
pub extern fn memcpy(dst: *mut c_void, src: *const c_void, size: size_t) {
	unsafe {
		calla0(0x2a, dst, src, size);
	}
}

#[no_mangle]
pub extern fn memset(dst: *mut c_void, chr: c_char, size: size_t) {
	unsafe {
		calla0(0x2b, dst, chr as c_int, size);
	}
}

//("This has a bug")
#[deprecated]
#[no_mangle]
pub extern fn memmove(dst: *mut c_void, src: *const c_void, size: size_t) {
	unsafe {
		calla0(0x2c, dst, src, size);
	}
}

//("This has a bug")
#[deprecated]
#[no_mangle]
pub extern fn memcmp(src1: *const c_void, src2: *const c_void, size: size_t) -> c_int {
	unsafe {
		calla0(0x2d, src1, src2, size) as c_int
	}
}

#[no_mangle]
pub extern fn memchr(src: *const c_void, chr: c_char, size: size_t) -> *const c_char {
	unsafe {
		calla0(0x2e, src, chr as c_int, size) as *const c_char
	}
}

#[no_mangle]
pub extern fn strlen(src: *const c_void) -> size_t {
	unsafe {
		calla0(0x1b, src)
	}
}
*/

// BIOS

#[cfg_attr(not(release), inline(never))]
fn puttychar(chr: u8) {
	unsafe {
		calla0(0x3c, chr as c_int);
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn putstty(s: &[u8]) {
	unsafe {
		calla0(0x3f, b"%s\n\0".as_ptr(), s.as_ptr());
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn set_default_rfe() {
	unsafe { callb0(0x18); }
}

/// BIOS: pad operations

#[derive(Copy, Clone)] // because it's just an u16
pub struct DigitalPadButtons{
	bits: u16,
}

impl DigitalPadButtons {
	pub fn new(bits: u16) -> DigitalPadButtons { DigitalPadButtons { bits: bits } }
	pub fn bits(&self) -> u16 { self.bits }

	pub fn select(&self)   -> bool { self.bits & (1 <<  0) != 0 }
	pub fn l3(&self)       -> bool { self.bits & (1 <<  1) != 0 }
	pub fn r3(&self)       -> bool { self.bits & (1 <<  2) != 0 }
	pub fn start(&self)    -> bool { self.bits & (1 <<  3) != 0 }
	pub fn up(&self)       -> bool { self.bits & (1 <<  4) != 0 }
	pub fn right(&self)    -> bool { self.bits & (1 <<  5) != 0 }
	pub fn down(&self)     -> bool { self.bits & (1 <<  6) != 0 }
	pub fn left(&self)     -> bool { self.bits & (1 <<  7) != 0 }
	pub fn l2(&self)       -> bool { self.bits & (1 <<  8) != 0 }
	pub fn r2(&self)       -> bool { self.bits & (1 <<  9) != 0 }
	pub fn l1(&self)       -> bool { self.bits & (1 << 10) != 0 }
	pub fn r1(&self)       -> bool { self.bits & (1 << 11) != 0 }
	pub fn triangle(&self) -> bool { self.bits & (1 << 12) != 0 }
	pub fn circle(&self)   -> bool { self.bits & (1 << 13) != 0 }
	pub fn cross(&self)    -> bool { self.bits & (1 << 14) != 0 }
	pub fn square(&self)   -> bool { self.bits & (1 << 15) != 0 }
}

impl Default for DigitalPadButtons {
	fn default() -> DigitalPadButtons { DigitalPadButtons{ bits: 0 } }
}

impl std::ops::BitXor for DigitalPadButtons {
	type Output = DigitalPadButtons;
	fn bitxor(self, other: DigitalPadButtons) -> DigitalPadButtons { DigitalPadButtons{ bits: self.bits ^ other.bits } }
}

impl std::ops::BitAnd for DigitalPadButtons {
	type Output = DigitalPadButtons;
	fn bitand(self, other: DigitalPadButtons) -> DigitalPadButtons { DigitalPadButtons{ bits: self.bits & other.bits } }
}

impl std::ops::Not for DigitalPadButtons {
	type Output = DigitalPadButtons;
	fn not(self) -> DigitalPadButtons { DigitalPadButtons{ bits: !self.bits } }
}

impl std::ops::BitOr for DigitalPadButtons {
	type Output = DigitalPadButtons;
	fn bitor(self, other: DigitalPadButtons) -> DigitalPadButtons { DigitalPadButtons{ bits: self.bits | other.bits } }
}

#[derive(Default)]
pub struct DigitalPadButtonState{
	prevbtns: DigitalPadButtons,
	btns: DigitalPadButtons,
}

impl DigitalPadButtonState {
	pub fn new() -> DigitalPadButtonState {
		DigitalPadButtonState { ..Default::default() }
	}
	pub fn update(&mut self, bits: u16) {
		self.prevbtns = self.btns;
		self.btns = DigitalPadButtons::new(bits);
	}

	pub fn press(&self) -> DigitalPadButtons {
		self.btns
	}

	pub fn down(&self) -> DigitalPadButtons {
		(self.prevbtns ^ self.btns) & self.btns
	}

	pub fn up(&self) -> DigitalPadButtons {
		(self.prevbtns ^ self.btns) & self.prevbtns
	}
}

// FIXME this goes bss... but can't we omit initial value from this source?
pub static mut padbuf: [[u8; 0x22]; 2] = [[0u8; 0x22], [0u8; 0x22]];

#[cfg_attr(not(release), inline(never))]
pub fn init_pad() {
	unsafe { callb0(0x12, padbuf[0].as_ptr(), 0x22, padbuf[1].as_ptr(), 0x22); }
}

#[cfg_attr(not(release), inline(never))]
pub fn start_pad() {
	unsafe { callb0(0x13); }
}

#[cfg_attr(not(release), inline(never))]
pub fn stop_pad() {
	unsafe { callb0(0x14); }
}

/// BIOS: event operations

// TODO how I quit specifying this overflowing literal?
pub enum EventClass {
	#[allow(overflowing_literals)]
	CDROM = 0xF0000003,
	#[allow(overflowing_literals)]
	VBlank = 0xF2000003, // not F0000001 for raw VBlank intr, that is erased by other BIOS handler
}

pub enum EventSpec {
	Interrupted = 2,
	Completed = 0x20,
	DataEnd = 0x80,
	Error = 0x8000,
}

pub enum EventMode {
	Func = 0x1000,
	NoFunc = 0x2000,
}

pub type EventHandle = u32;

#[cfg_attr(not(release), inline(never))]
pub fn open_event(class: EventClass, spec: EventSpec, opfunc: Option<&Fn()>) -> Option<EventHandle> {
	let r = if let Some(func) = opfunc {
		unsafe { callb0(0x08, class as u32, spec as u32, EventMode::Func, func as *const Fn()) }
	} else {
		unsafe { callb0(0x08, class as u32, spec as u32, EventMode::NoFunc, 0) }
	};

	if r as i32 != -1 {
		Some(r as EventHandle)
	} else {
		None
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn enable_event(event: EventHandle) {
	unsafe { callb0(0x0c, event); }
}

#[cfg_attr(not(release), inline(never))]
pub fn wait_event(event: EventHandle) {
	unsafe { callb0(0x0a, event); }
}

#[cfg_attr(not(release), inline(never))]
pub fn test_event(event: EventHandle) -> bool {
	match unsafe { callb0(0x0b, event) } {
		0 => false, // not ready
		1 => true, // ready
		r => panic!("Unexpected TestEvent result: {}", r)
	}
}

/// BIOS: file operations
// note: some functions are in both a0 and b0 syscalls, but b0 are in main memory,
//       may be faster than a0 (that in the ROM that is narrower bus.)

pub struct FileDescr { fd: u32 }

pub type FileMode = u32; // u16, avoid casting...
pub enum ReadWrite {
	Read,
	Write
}
// FIXME mod+consts looks bad...
pub mod file_mode {
	pub const READ: super::FileMode = 1 << 0; // specified by internal code
	pub const WRITE: super::FileMode = 1 << 1; // specified by internal code
	pub const ASYNC_TTY: super::FileMode = 1 << 2; // specified by user
	pub const CREATE: super::FileMode = 1 << 9; // specified by internal code
	pub const SYNC: super::FileMode = 0; // specified by user, dummy.
	pub const ASYNC: super::FileMode = 1 << 15; // specified by user. TODO get its completion by event (class=filedescr,spec=4)?
}

pub enum SeekType {
	Set = 0,
	Curr = 1,
	#[deprecated]
	NoOp = 2 // Should "End", but do nothing in SCPH1001
}

// TODO specifying concrete values for expecting something .from_int, but no such.
//      so values are optional. from_int-ish is now get_last_error.
#[derive(Copy, Clone, Debug, PartialEq)] // because this is just an integer.
pub enum Errno {
	Ok = 0x00,
	FileNotFound = 0x02,
	BadDevicePort = 0x06,
	InvalidHandle = 0x09,
	GeneralError = 0x10,
	AlreadyExists = 0x11,
	RenameNotOnSameDevice = 0x12,
	UnknownDevice = 0x13,
	AlignmentError = 0x16, // or exceeding filesize
	HandleSpaceFull = 0x18,
	MemoryBlockFull = 0x1C,

	ProgramError = 0xFE, // generated by this program, not BIOS.
}

impl Default for Errno {
	fn default() -> Errno { Errno::Ok }
}

#[repr(C)]
#[derive(Default, Clone)] // FIXME Clone is not intended
pub struct Dirent {
	pub name: [u8; 0x14],
	pub attr: u32,
	pub size: u32,
	pub nextent: u32, // not used? (NO$)
	pub firstsector: u32,
	pub reserved: u32 // not used (NO$)
}

impl<'a> Dirent {
#[cfg_attr(not(release), inline(never))]
	pub fn name(&'a self) -> &'a[u8] {
		&self.name[0..self.name.iter().take_while(|c|**c != 0).count()]
	}
}

// should say &unsafe Fn()... match unsafe{fun()} ..., but Rust 1.10 does not accept.
// FIXME taking doop as &FnMut is ugly, but
//       * I want to do mut-ing some value in that in some situation,
//       * I want to make wrap_filecall as a single function, not monomorphizated functions.
#[cfg_attr(not(release), inline(never))]
fn wrap_filecall<T>(doop: &mut FnMut() -> usize, badvalue: i32, convop: &Fn(usize) -> T) -> Result<T, Errno> {
	let r = doop();
	if r as i32 == badvalue {
		Result::Err(get_last_error())
	} else {
		Result::Ok(convop(r))
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn get_last_error() -> Errno {
	unsafe {
		match callb0(0x54) {
			0x00 => Errno::Ok,
			0x02 => Errno::FileNotFound,
			0x06 => Errno::BadDevicePort,
			0x09 => Errno::InvalidHandle,
			0x10 => Errno::GeneralError,
			0x11 => Errno::AlreadyExists,
			0x12 => Errno::RenameNotOnSameDevice,
			0x13 => Errno::UnknownDevice,
			0x16 => Errno::AlignmentError,
			0x18 => Errno::HandleSpaceFull,
			0x1C => Errno::MemoryBlockFull,
			errno => panic!("Unknown Errno {:02X}", errno)
		}
	}
}

#[inline(always)]
fn filename_asserts(filename: &[u8]) {
	// this restricts only for name without device/path, and only for memcard...
	//assert!(filename.len() <= 21, "filename must be shorter than 20 bytes (without null-termination)");
	assert!(filename[filename.len() - 1] == 0, "filename must be null-terminated");
}

// note: filename must be self null-term!
#[cfg_attr(not(release), inline(never))]
fn file_open_intern(filename: &[u8], rw: ReadWrite, mode: FileMode) -> Result<FileDescr, Errno> {
	filename_asserts(filename);
	assert!(mode & (file_mode::READ | file_mode::WRITE) == 0, "mode must not contains READ/WRITE, as specified by ReadWrite");
	let mode = mode | match rw {
		ReadWrite::Read => file_mode::READ,
		ReadWrite::Write => file_mode::WRITE
	};
	return wrap_filecall(&mut || unsafe{callb0(0x32, filename.as_ptr(), mode)}, -1, &|r| FileDescr{fd: r as u32});
}

// FIXME mut-ing ref of FileDescr makes(?) allocation in stack and emit load/store that is unnecessary.

#[cfg_attr(not(release), inline(never))]
pub fn file_seek(fd: &mut FileDescr, offset: i32, seektype: SeekType) -> Result<usize, Errno> {
	let st32 = seektype as u32; // FIXME ugly...
	return wrap_filecall(&mut || unsafe{callb0(0x33, fd.fd, offset, st32)}, -1, &|r| r);
}

#[cfg_attr(not(release), inline(never))]
pub fn file_read(fd: &mut FileDescr, buf: &mut[u8]) -> Result<usize, Errno> {
	return wrap_filecall(&mut || unsafe{callb0(0x34, fd.fd, buf.as_ptr(), buf.len())}, -1, &|r| r);
}

#[cfg_attr(not(release), inline(never))]
pub fn file_write(fd: &mut FileDescr, buf: &[u8]) -> Result<usize, Errno> {
	return wrap_filecall(&mut || unsafe{callb0(0x35, fd.fd, buf.as_ptr(), buf.len())}, -1, &|r| r);
}

/*
#[cfg_attr(not(release), inline(never))]
pub fn file_close(fd: FileDescr) -> Result<(), Errno> {
	return wrap_filecall(&|| unsafe{callb0(0x36, fd.fd)}, -1, &|_| ());
}
*/

impl std::ops::Drop for FileDescr {
#[cfg_attr(not(release), inline(never))]
	fn drop(&mut self) {
		//file_close(self);
		return wrap_filecall(&mut || unsafe{callb0(0x36, self.fd)}, -1, &|_| ()).expect("close failed!?");
	}
}

// XXX WARNING: DO NOT USE FOR cdrom:*, NO$PSX 1.9 BIOS' first_file(b0_42) has a critical bug that does not return from it at all!!
#[cfg_attr(not(release), inline(never))]
pub fn first_file(filename: &[u8]) -> Result<Dirent, Errno> {
	filename_asserts(filename);
	// FIXME manually mapping is ugly... but ent can't be &mut and move in a single function call.
	// XXX umm... no need to initialize, overkill...
	let mut ent = Default::default();
	let r = wrap_filecall(&mut || unsafe{callb0(0x42, filename.as_ptr(), &mut ent as *mut Dirent)}, 0, &|_| ());
	return r.map(|_| ent);
}

// note: DO NOT call next_file after opening some file, due to a bug in BIOS!
//       (instead, retry from first_file.)
#[cfg_attr(not(release), inline(never))]
pub fn next_file(ent: &mut Dirent) -> Result<(), Errno> {
	return wrap_filecall(&mut || unsafe{callb0(0x43, ent as *mut Dirent)}, 0, &|_| ());
}

#[cfg_attr(not(release), inline(never))]
pub fn file_delete(filename: &[u8]) -> Result<(), Errno> {
	filename_asserts(filename);
	return wrap_filecall(&mut || unsafe{callb0(0x45, filename.as_ptr())}, 0, &|_| ());
}

#[cfg_attr(not(release), inline(never))]
pub fn format_device(devname: &str) -> Result<(), Errno> {
	return wrap_filecall(&mut || unsafe{callb0(0x41, devname.as_ptr())}, 0, &|_| ());
}

/// BIOS: font functions

// note: returned bitmap is big-endian, so can't use u16.
// FIXME should ->Option<&'static[u16;15]>
#[cfg_attr(not(release), inline(never))]
pub fn sjis2ptr(chr: u16) -> Option<&'static[u8]> {
	let ptr = unsafe { callb0(0x51, chr as u32) };
	if ptr == 0xFFFFffff {
		return Option::None;
	}
	return Option::Some(unsafe { std::slice::from_raw_parts(ptr as *const u8, 2 * 15) }); // 16px wide(2bytes), 15px height
}

/// BIOS: critical section functions

//#[cfg_attr(not(release), inline(never))]
pub fn enter_critsect() {
	unsafe { syscall(0x01); }
}

//#[cfg_attr(not(release), inline(never))]
pub fn leave_critsect() {
	unsafe { syscall(0x02); }
}

/// BIOS: raw CD reading functions

pub struct CDError {}
impl fmt::Debug for CDError {
	fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
		write!(fmt, "<CDError>")
	}
}
impl fmt::Display for CDError {
	fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
		write!(fmt, "<CDError>")
	}
}
impl Error for CDError {
	fn description(&self) -> &str {
		"Just a CD function failure"
	}
}

macro_rules! cd_func_wrap {
	($name:expr, $($args:expr),+) => {
		match unsafe { calla0($($args),+) } {
			0 => Result::Err(CDError{}),
			1 => Result::Ok(()),
			r => panic!(concat!("Unexpected ", $name, " result: {}"), r),
		}
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn cd_async_seekl(lba: u32) -> Result<(), CDError> {
	let sector = lba + 150;
	let msf = [(sector / 75 / 60) as u8, (sector / 75 % 60) as u8, (sector % 75) as u8];
	cd_func_wrap!("CdAsyncSeekL", 0x78, &msf as *const u8)
}

#[cfg_attr(not(release), inline(never))]
pub fn cd_async_get_status(dst: &mut u8) -> Result<(), CDError> {
	cd_func_wrap!("CdAsyncSeekL", 0x78, dst as *mut u8)
}

#[cfg_attr(not(release), inline(never))]
pub fn cd_async_read_sector(count: u32, dst: &mut [u8], mode: u32) -> Result<(), CDError> {
	assert!(2048 * count as usize <= dst.len(), "dst too small");
	cd_func_wrap!("CdAsyncReadSector", 0x7E, count, dst.as_mut_ptr(), mode)
}

#[cfg_attr(not(release), inline(never))]
pub fn cd_get_status() -> Result<u8, CDError> {
	match unsafe { calla0(0xA6) as i32 } {
		r @ 0...0xFF => Result::Ok(r as u8),
		-1 => Result::Err(CDError{}),
		r => panic!("Unexpected CdGetStatus result: {}", r as u32)
	}
}

#[cfg_attr(not(release), inline(never))]
pub fn cd_read_sector(count: u32, sector: u32, dst: &mut [u8]) -> Result<u32, CDError> {
	assert!(2048 * count as usize <= dst.len(), "dst too small");
	let r = unsafe { calla0(0xA5, count, sector, dst.as_mut_ptr()) as i32 };
	if 0 <= r {
		Result::Ok(r as u32)
	} else if r == -1 {
		Result::Err(CDError{})
	} else {
		panic!("Unexpected CdReadSector status: {}", r as u32)
	}
}

/// BIOS: memcard functions

#[cfg_attr(not(release), inline(never))]
pub fn init_card(pad_enable: bool) {
	unsafe { callb0(0x4A, if pad_enable { 1 } else { 0 }); }
}

#[cfg_attr(not(release), inline(never))]
pub fn start_card() {
	unsafe { callb0(0x4B); }
}

#[cfg_attr(not(release), inline(never))]
pub fn stop_card() {
	unsafe { callb0(0x4C); }
}

#[cfg_attr(not(release), inline(never))]
pub fn bu_init() {
	unsafe { calla0(0x55); }
}

// BIOS support function

#[cfg_attr(not(release), inline(never))]
pub fn printty(s: &str) {
	for c in s.chars() {
		puttychar(c as u8);
	}
}

// note: filename must be self null-term!
#[cfg_attr(not(release), inline(never))]
pub fn file_open(filename: &[u8], rw: ReadWrite, mode: FileMode) -> Result<FileDescr, Errno> {
	assert!(mode & file_mode::CREATE == 0, "mode must not contains CREATE, use file_create");
	return file_open_intern(filename, rw, mode as u32);
}

// note: filename must be self null-term!
#[cfg_attr(not(release), inline(never))]
pub fn file_create(filename: &[u8], mode: FileMode, blocks: u16) -> Result<FileDescr, Errno> {
	assert!(mode & file_mode::CREATE == 0, "mode must not contains CREATE, as this function name already says");
	return file_open_intern(filename, ReadWrite::Write, mode | file_mode::CREATE | ((blocks as u32) << 16));
}

impl io::Read for FileDescr {
#[cfg_attr(not(release), inline(never))]
	fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
		return file_read(self, buf).or(Result::Err(io::Error::new(io::ErrorKind::Other, "something happened")));
	}
}

pub struct DirFinder {
	filename: Vec<u8>,
	dirent: Option<Dirent>,
	// TODO embedding lasterror in an iterator is useless, as #map etc. eats it.
	//      use raw first_file/next_file and user manually call get_last_error.
	lasterror: Errno
}

impl DirFinder {
#[cfg_attr(not(release), inline(never))]
	pub fn new(filename: &[u8]) -> DirFinder {
		filename_asserts(filename);
		DirFinder { filename: filename.to_vec(), dirent: Option::None, lasterror: Errno::Ok }
	}

#[cfg_attr(not(release), inline(never))]
	pub fn get_last_error(&self) -> Errno {
		return self.lasterror;
	}
}

impl Iterator for DirFinder {
	type Item = Dirent; // FIXME no clone (value itself) but ref...

#[cfg_attr(not(release), inline(never))]
	fn next(&mut self) -> Option<Dirent> {
		let r = match self.dirent {
			// first iterate
			None =>
				first_file(&self.filename).and_then(|ent| {
					self.dirent = Option::Some(ent); // lost virgin
					Result::Ok(())
				}),
			// second or later iterate
			Some(ref mut dirent) =>
				next_file(dirent),
		};
		return match r {
			Ok(_) => self.dirent.clone(),
			Err(errno) => {
				// TODO first_file/next_file returns 0 if no file found, what is the errno?
				//      ->it seems 9(invalidhandle)
				self.lasterror = errno;
				Option::None
			}
		};
	}
}

// FIXME quick hack!!
impl std::convert::From<io::Error> for Errno {
	fn from(_: io::Error) -> Errno {
		return Errno::ProgramError;
	}
}

// ex. write_bu(b"bu00:BISXXS-XXXXXfoo", sys::FileMode::SYNC, 1, sjfw("Hi"), (pal, pixmaps), |mut fd|{ sys::file_write(fd, buf); })
// note: does not support pocket-station area.
#[cfg_attr(not(release), inline(never))]
pub fn write_bu<T, F>(
		filename: &[u8], // must be null-term
		mode: FileMode,
		nblocks: u32,
		title: &[u16],
		icon: (&[u16;16],&[[u8;128]]),
		f: F
		) -> Result<T, Errno>
		where F: FnOnce(&mut FileDescr) -> Result<T, Errno> {
	assert!(1 <= nblocks && nblocks <= 15, "nblocks must be 1...15");
	// title: really 64bytes, but must <=32 all-fullwidth for (at least) SCPH1001 BIOS memory card menu, otherwise it work buggy.
	assert!(title.len() <= 32, "title must be in 32 or less characters");
	assert!(1 <= icon.1.len() && icon.1.len() <= 3, "icon pixmap must be 1...3 frame(s)");

	let mut fd = try!(file_create(filename, mode, nblocks as u16).or_else(|err| {
		if err == Errno::AlreadyExists {
			// TODO should check nblocks and reuse? but how to get nblocks?
			file_delete(filename).and_then(|_| {
				file_create(filename, mode, nblocks as u16)
			})
		} else {
			Result::Err(err)
		}
	}));

	let mut hdr: Vec<u8> = Vec::with_capacity(128);

	// magic, icontype(frames), nblocks
	try!(hdr.write(&[b'S', b'C', (0x10 + icon.1.len()) as u8, nblocks as u8]));
	// title
	{
		let mut tb = [0u8; 64];
		let mut i = 0;
		for c in title {
			// MSB first in SJIS
			// TODO are hw-chars allowed? if not, this "if" is not necessary. always push *c>>8.
			if 0x100 <= *c {
				tb[i] = (*c >> 8) as u8;
				i += 1;
			}
			tb[i] = (*c & 0xFF) as u8;
			i += 1;
		}
		try!(hdr.write(&tb));
	}
	// reserved
	try!(hdr.write(&[0; 0x60-0x44]));
	// icon-clut
	try!(hdr.write(&icon.0.iter().fold(Vec::with_capacity(32), |mut s,e| {
		s.push((*e & 0xFF) as u8);
		s.push((*e >> 8) as u8);
		s
	}))); // TODO transmute?
	// write a (header) sector
	assert!(hdr.len() == 128, "header (without icon pixmaps) must be 128 bytes here.");
	try!(file_write(&mut fd, &hdr));
	// write icon pixmaps
	for ifr in icon.1 { try!(file_write(&mut fd, ifr)); }
	// do anything (write body)
	let ret = f(&mut fd);
	// close by let fd scope out
	return ret;
}

pub fn in_critical_section<T>(f: T) where T: FnOnce() {
	enter_critsect();
	f();
	leave_critsect();
}

#[cfg_attr(not(release), inline(never))]
pub fn cd_read(sector: u32, count: u32) -> Result<Vec<u8>, CDError> {
	let mut buf = Vec::with_capacity(2048 * count as usize);
	unsafe { buf.set_len(2048 * count as usize); } // FIXME init not required and it is vec of u8 that is !Drop, but "unsafe"...
	match cd_read_sector(count, sector, &mut buf) {
		Ok(r) => {
			buf.truncate(2048 * r as usize);
			Result::Ok(buf)
		},
		Err(e) => Result::Err(e)
	}
}
