use ::{Scene, GraphicLogger, gpu_setres_raw, gp0, gp1, fillvram, draw_ascii_text_cached};
use ::sys::{cd_read, DigitalPadButtonState, DigitalPadButtons};
use ::fastmath::{sin, cos, Qang, Q};
use ::mat::{load_mat9_rt, load_mat9_ll, load_mat9_lc, mult_load_mat9_rt, mult_load_mat9_ll};
use ::graphic::{TDModel, OrderTab};
use std::default::Default;

pub struct TD1Scene {
	frame: u32,
	model: TDModel,
	ot_depth: usize,
	btns: DigitalPadButtonState,
	btnrepeat: u32,
	rot: [i32; 3],
	tr: [i32; 3],
	h: u16,
	sz: i16,
	lumi: bool,
	ll: bool,
}

impl TD1Scene {
	#[cfg_attr(not(release), inline(never))]
	pub fn new(_logger: &mut GraphicLogger) -> TD1Scene {
		let model = TDModel::load(cd_read(5, 7).expect("could not read PSLogo"));
		gpu_setres_raw(1, 0, 0, 0);
		TD1Scene {
			frame: Default::default(),
			model: model,
			ot_depth: 1024,
			btns: Default::default(),
			btnrepeat: Default::default(),
			rot: Default::default(),
			tr: [0, 0, 0x1700],
			h: 0x400,
			sz: 0x1000,
			lumi: true,
			ll: false,
		}
	}
}

impl Scene for TD1Scene {
	#[cfg_attr(not(release), inline(never))]
	#[allow(overflowing_literals)] // FIXME mtc2
	fn draw(&self) {
		// switch displaying frame
		let preyoff = self.frame * 240;
		gp1(0x05000000 | (  0 << 0) | (preyoff << 10));
		// setup
		gp0(0xE1000000); // texpage (almost for non-texturedpolys)
		gp0(0xE2000000); // texwindow
		gp0(0xE6000000); // maskbits
		// switch drawing frame
		let curyoff = (1 - self.frame) * 240;
		gp0(0xE3000000 | (  0 << 0) | (curyoff << 10)); // draw(clip)x1y1
		gp0(0xE4000000 | (319 << 0) | ((curyoff + 239) << 10)); // draw(clip)x2y2
		gp0(0xE5000000 | (  0 << 0) | (curyoff << 11)); // drawoffset
		// fillvram is not affected by drawoffset...
		fillvram(64, 64, 64, 0, curyoff as u16, 320, 240);

		// prepare OT
		let mut ot = OrderTab::new(self.ot_depth, 5*300/*just guess*/);

		// calc LC for lighting
		{
			// initial
			load_mat9_lc(&[0, 0, 0x1000, 0, 0, 0x1000, 0, 0, 0x1000]);
		}

		// calc mat that rotates by y(yaw),x(pitch),z(roll).
		{
			// initial
			load_mat9_rt(&[0x1000, 0, 0, 0, 0x1000, 0, 0, 0, 0x1000]);
			load_mat9_ll(&[0x1000, 0, 0, 0, 0x1000, 0, 0, 0, 0xF000]);

			// scale z
			mult_load_mat9_rt(&[
								 0x1000, 0, 0,
								 0, 0x1000, 0,
								 0, 0, self.sz,
			]);
			/*
			if self.ll {
				mult_load_mat9_ll(&[
									 0x1000, 0, 0,
									 0, 0x1000, 0,
									 0, 0, self.sz,
				]);
			}
			*/

			// calculate z
			{
				let theta = Qang::from_deg(self.rot[2]);
				let c = cos(theta).raw_i16();
				let s = sin(theta).raw_i16();
				mult_load_mat9_rt(&[
								 c, s, 0,
								 -s, c, 0,
								 0, 0, 0x1000,
				]);
				if self.ll {
					mult_load_mat9_ll(&[
									 c, s, 0,
									 -s, c, 0,
									 0, 0, 0x1000,
					]);
				}
			}

			// calculate (z)*x
			{
				let theta = Qang::from_deg(self.rot[0]);
				let c = cos(theta).raw_i16();
				let s = sin(theta).raw_i16();
				mult_load_mat9_rt(&[
								 0x1000, 0, 0,
								 0, c, -s,
								 0, s, c,
				]);
				if self.ll {
					mult_load_mat9_ll(&[
									 0x1000, 0, 0,
									 0, c, -s,
									 0, s, c,
					]);
				}
			}

			// calculate (z*x)*y
			{
				let theta = Qang::from_deg(self.rot[1]);
				let c = cos(theta).raw_i16();
				let s = sin(theta).raw_i16();
				mult_load_mat9_rt(&[
									 c, 0, s,
									 0, 0x1000, 0,
									 -s, 0, c,
				]);
				if self.ll {
					mult_load_mat9_ll(&[
										 c, 0, s,
										 0, 0x1000, 0,
										 -s, 0, c,
					]);
				}
			}
		}
		mtc2!(37, self.tr[0]); mtc2!(38, self.tr[1]); mtc2!(39, self.tr[2]);

		// matrix and vectors
		//mtc2!(32, 0x00000FFB); mtc2!(33, 0xFFB7FF44); mtc2!(34, 0xF9CA0EBC); mtc2!(35, 0x063700AD); mtc2!(36, 0x0EB7); // RT3x3q12
		//mtc2!(40, 0x00000000); mtc2!(41, 0x00000FA0); mtc2!(42, 0x0000F060); mtc2!(43, 0x0000F060); mtc2!(44, 0x0000); // LM3x3q12
		//mtc2!(40, 0x00001000); mtc2!(41, 0x00000000); mtc2!(42, 0x00001000); mtc2!(43, 0x00000000); mtc2!(44, 0x1000); // LM3x3q12
		//mtc2!(40, 0x00001000); mtc2!(41, 0x00000000); mtc2!(42, 0x00001000); mtc2!(43, 0x00000000); mtc2!(44, 0xF000); // LM3x3q12
		/*
		if self.lumi {
			mtc2!(48, 0x0BB80FA0); mtc2!(49, 0x0FA00FA0); mtc2!(50, 0x0FA00BB8); mtc2!(51, 0x0BB80FA0); mtc2!(52, 0x0FA0); // LL3x3q12
		} else {
			mtc2!(48, 0x10001000); mtc2!(49, 0x10001000); mtc2!(50, 0x10001000); mtc2!(51, 0x10001000); mtc2!(52, 0x1000); // LL3x3q12
		}
		*/
		//mtc2!(37, 0x00000000); mtc2!(38, 0x00000000); mtc2!(39, 0x00001700); // TRq0 0,-170,5888
		//mtc2!(45, 0x00000640); mtc2!(46, 0x00000640); mtc2!(47, 0x00000640); // BKq12?
		mtc2!(45, 0x00000000); mtc2!(46, 0x00000000); mtc2!(47, 0x00000000); // BKq12?
		mtc2!(53, 0x00000000); mtc2!(54, 0x00000000); mtc2!(55, 0x00000000); // FCq4
		// for RTP
		mtc2!(56, 0x00A00000); mtc2!(57, 0x00780000); // OFX,OFYq16 160,120
		//mtc2!(58, 0x0400); // Hq0u16
		mtc2!(58, self.h); // Hq0u16
		mtc2!(59, 0xFC82); // DQAq8 (coeff)
		mtc2!(60, 0x01400000); // DQBq24? (offset)
		// for AVSZ
		mtc2!(61, 0x0155>>1); // ZSF3q12 (1/3 / 2)
		mtc2!(62, 0x0100>>1); // ZSF4q12 (1/4 / 2)

		// settle OT before inserting
		ot.settle();
		ot.enqueue(&self.model);
		ot.submit();
		ot.wait(); // required on XEBRA 150831/ePSXe 2.0.5
		{
			let mut y = 16;
			let diag = ot.get_diag().expect("Not debug!?");
			draw_ascii_text_cached(0, y, 0x10000, 0x10000, &format!("{} polys", diag.npoly).as_bytes()); y += 8;
			draw_ascii_text_cached(0, y, 0x10000, 0x10000, &format!("r={},{},{}", self.rot[0], self.rot[1], self.rot[2]).as_bytes()); y += 8;
			draw_ascii_text_cached(0, y, 0x10000, 0x10000, &format!("t={},{},{}", self.tr[0], self.tr[1], self.tr[2]).as_bytes()); y += 8;
			draw_ascii_text_cached(0, y, 0x10000, 0x10000, &format!("h={}", self.h).as_bytes()); y += 8;
			draw_ascii_text_cached(0, y, 0x10000, 0x10000, &format!("sz={}", self.sz).as_bytes()); y += 8;
			draw_ascii_text_cached(0, y, 0x10000, 0x10000, &format!("z={}..{}", diag.zrange.0, diag.zrange.1).as_bytes()); y += 8;
		}
	}
	#[cfg_attr(not(release), inline(never))]
	fn input(&mut self, bits: u16) {
		self.btns.update(bits);
		let press = self.btns.press();
		self.btnrepeat = if press.bits() & 0x00F0 == 0 { 0 } else { self.btnrepeat + 1 };
		let down = self.btns.down() | if 18 < self.btnrepeat { press } else { DigitalPadButtons::new(0) };

		let delta = if press.l1() || press.l2() { 16 } else { 1 };
		if press.cross() {
			if down.left() {
				self.rot[2] -= delta;
			} else if down.right() {
				self.rot[2] += delta;
			}
			if down.up() {
				self.tr[2] += delta;
			} else if down.down() {
				self.tr[2] -= delta;
			}
		} else if press.square() {
			if down.left() {
				self.rot[1] -= delta;
			} else if down.right() {
				self.rot[1] += delta;
			}
			if down.up() {
				self.rot[0] -= delta;
			} else if down.down() {
				self.rot[0] += delta;
			}
		} else if press.circle() {
			if down.left() { // umm
				self.sz -= delta as i16;
			} else if down.right() {
				self.sz += delta as i16;
			}
			if down.up() {
				self.h -= delta as u16;
			} else if down.down() {
				self.h += delta as u16;
			}
		} else {
			if down.left() {
				self.tr[0] -= delta;
			} else if down.right() {
				self.tr[0] += delta;
			}
			if down.up() {
				self.tr[1] -= delta;
			} else if down.down() {
				self.tr[1] += delta;
			}
		}

		if down.select() {
			self.ll = !self.ll;
		}
		if down.start() {
			self.lumi = !self.lumi;
		}

		self.frame = (self.frame + 1) % 2;
	}
}
