use std::intrinsics::{volatile_load, volatile_store};
use super::sys;
use super::{GraphicLogger, cache_ascii_font, CACHED_FONT_CW, CACHED_FONT_CH, draw_ascii_text_cached, fillrect, gp1disprange, calc_gpu_res_params, gp1, Scene, gpustat, gp0, fillvram, binread, spu_dmaload, spu_play, ascii2sjis};

#[derive(Default)]
pub struct TestScene {
	btns: sys::DigitalPadButtonState,
	parx: usize,
	pary: usize,
	romfiles: Option<(Vec<Vec<u8>>, sys::Errno)>,
	bufiles: Option<(Vec<Vec<u8>>, sys::Errno)>,
	cnf: Option<Result<Vec<u8>, sys::Errno>>,
	wrerr: Option<u32>,
	//fields: Vec<Box<FnMut(bool, sys::DigitalPadButtons)>>,
	fields: Vec<Box<TestSceneField>>,
	frame: u32,
	gpupars: [u32; 4],
	deg: i32,
}

trait TestSceneField {
	fn draw(&self, ts: &TestScene, active: bool);
	// FIXME *mut TestScene... DO NOT TRY THIS AT HOME!
	fn input(&mut self, ts: *mut TestScene, btns: &sys::DigitalPadButtonState);
	fn columns(&self) -> usize;
}

impl TestScene {
#[cfg_attr(not(release), inline(never))]
	pub fn new(logger: &mut GraphicLogger) -> TestScene {
		const TSX: u32 = 0x10000;
		const TSY: u32 = 0x10000;
		const TW: u16 = (CACHED_FONT_CW as u32 * TSX / 0x10000) as u16;
		const TH: u16 = (CACHED_FONT_CH as u32 * TSY / 0x10000) as u16;

		let par1xy = (16, 48);
		let par2xy = (16 + TW * 13, 48);
		let par3xy = (16 + TW * 17, 48 + TH * 12);

		let mut fields: Vec<Box<TestSceneField>> = Vec::new();

		// one-halfword SPU reverb params
		{
			struct OneReverb {
				ptr: *mut u16,
				pos: (u16, u16),
				name: &'static [u8],
			}
			impl TestSceneField for OneReverb {
				fn draw(&self, ts: &TestScene, active: bool) {
					let pval = unsafe { volatile_load(self.ptr) };
					draw_ascii_text_cached(self.pos.0, self.pos.1, TSX, TSY, self.name);
					if active {
						fillrect(0, 192, 0, self.pos.0 + (7 + ts.parx as u16) * TW, self.pos.1, TW, TH);
					}
					draw_ascii_text_cached(self.pos.0 + 7 * TW, self.pos.1, TSX, TSY, &format!("{:04X}", pval).bytes().collect::<Vec<u8>>());
				}
				fn input(&mut self, pts: *mut TestScene, btns: &sys::DigitalPadButtonState) {
					let ts = unsafe { pts.as_mut().unwrap() };
					let change = |f: &Fn(i32, i32) -> i32| {
						let oval = unsafe { volatile_load(self.ptr) };
						let sh = 12 - (ts.parx * 4);
						let nval = f(oval as i32, 1 << sh) as u16;
						unsafe { volatile_store(self.ptr, nval); }
					};
					if btns.down().up() {
						change(&|v,d|v+d);
					}
					if btns.down().down() {
						change(&|v,d|v-d);
					}
					if btns.down().triangle() {
						unsafe { volatile_store(self.ptr, 0); }
					}
				}
				fn columns(&self) -> usize {
					4
				}
			}
			fields.push(Box::new(OneReverb {
					ptr: 0x1F801DA2 as *mut u16,
					pos: (par1xy.0, par1xy.1),
					name: &b"mBase"[..],
				}));
			for (i, n) in [&b"dApf1"[..], &b"dApf2"[..], &b"vIir"[..], &b"vComb1"[..], &b"vComb2"[..], &b"vComb3"[..], &b"vComb4"[..], &b"vWall"[..], &b"vApf1"[..], &b"vApf2"[..]].into_iter().enumerate() {
				fields.push(Box::new(OneReverb {
					ptr: (0x1F801DC0 + i * 2) as *mut u16,
					pos: (par1xy.0, par1xy.1 + (1 + i as u16) * TH),
					name: n,
				}));
			}
		}
		// two-halfwords SPU reverb params
		{
			struct TwoReverb {
				ptr: *mut u16,
				pos: (u16, u16),
				name: &'static [u8],
			}
			impl TestSceneField for TwoReverb {
				fn draw(&self, ts: &TestScene, active: bool) {
					let pvall = unsafe { volatile_load(self.ptr) };
					let pvalr = unsafe { volatile_load(self.ptr.offset(1)) };
					draw_ascii_text_cached(self.pos.0, self.pos.1, TSX, TSY, self.name);
					if active {
						let x = (ts.parx + if 3 < ts.parx { 1 } else { 0 }) as u16;
						fillrect(0, 192, 0, self.pos.0 + (7 + x) * TW, self.pos.1, TW, TH);
					}
					draw_ascii_text_cached(self.pos.0 +  7 * TW, self.pos.1, TSX, TSY, &format!("{:04X}", pvall).bytes().collect::<Vec<u8>>());
					draw_ascii_text_cached(self.pos.0 + 12 * TW, self.pos.1, TSX, TSY, &format!("{:04X}", pvalr).bytes().collect::<Vec<u8>>());
				}
				fn input(&mut self, pts: *mut TestScene, btns: &sys::DigitalPadButtonState) {
					let ts = unsafe { pts.as_mut().unwrap() };
					let change = |f: &Fn(i32, i32) -> i32| {
						let ppar = unsafe { self.ptr.offset(ts.parx as isize / 4) };
						let oval = unsafe { volatile_load(ppar) };
						let sh = 12 - (ts.parx % 4 * 4);
						let nval = f(oval as i32, 1 << sh) as u16;
						unsafe { volatile_store(ppar, nval); }
					};
					if btns.down().up() {
						change(&|v,d|v+d);
					}
					if btns.down().down() {
						change(&|v,d|v-d);
					}
					if btns.down().triangle() {
						unsafe {
							volatile_store(self.ptr, 0);
							volatile_store(self.ptr.offset(1), 0);
						}
					}
				}
				fn columns(&self) -> usize {
					8
				}
			}
			for (i, n) in [&b"mSame"[..], &b"mComb1"[..], &b"mComb2"[..], &b"dSame"[..], &b"mDiff"[..], &b"mComb3"[..], &b"mComb4"[..], &b"dDiff"[..], &b"mApf1"[..], &b"mApf2"[..], &b"vIn"[..]].iter().enumerate() {
				fields.push(Box::new(TwoReverb {
					ptr: (0x1F801DD4 + i * 4) as *mut u16,
					pos: (par2xy.0, par2xy.1 + i as u16 * TH),
					name: n,
				}));
			}
		}
		// two-halfwords GPU display params
		{
			struct TwoGPUDisp {
				idx: usize,
				pos: (u16, u16),
				name: &'static [u8],
			}
			impl TestSceneField for TwoGPUDisp {
				fn draw(&self, ts: &TestScene, active: bool) {
					let pvall = ts.gpupars[self.idx * 2 + 0] as u16;
					let pvalr = ts.gpupars[self.idx * 2 + 1] as u16;
					draw_ascii_text_cached(self.pos.0, self.pos.1, TSX, TSY, &self.name);
					if active {
						let x = (ts.parx + if 3 < ts.parx { 1 } else { 0 }) as u16;
						fillrect(0, 192, 0, self.pos.0 + (7 + x) * TW, self.pos.1, TW, TH);
					}
					draw_ascii_text_cached(self.pos.0 +  7 * TW, self.pos.1, TSX, TSY, &format!("{:04X}", pvall).bytes().collect::<Vec<u8>>());
					draw_ascii_text_cached(self.pos.0 + 12 * TW, self.pos.1, TSX, TSY, &format!("{:04X}", pvalr).bytes().collect::<Vec<u8>>());
				}
				fn input(&mut self, pts: *mut TestScene, btns: &sys::DigitalPadButtonState) {
					let ts = unsafe { pts.as_mut().unwrap() };
					let mut change = |f: &Fn(i32, i32) -> i32| {
						let i = self.idx * 2 + if 3 < ts.parx { 1 } else { 0 };
						let sh = 12 - (ts.parx % 4 * 4);
						ts.gpupars[i] = f(ts.gpupars[i] as i32, 1 << sh) as u32;
						gp1disprange(ts.gpupars[0], ts.gpupars[1], ts.gpupars[2], ts.gpupars[3]);
					};
					if btns.down().up() {
						change(&|v,d|v+d);
					}
					if btns.down().down() {
						change(&|v,d|v-d);
					}
					if btns.down().triangle() {
						// not supported
					}
				}
				fn columns(&self) -> usize {
					8
				}
			}
			for (i, n) in [&b"horz"[..], &b"vert"[..]].iter().enumerate() {
				fields.push(Box::new(TwoGPUDisp {
					idx: i,
					pos: (par1xy.0, par1xy.1 + (12 + i) as u16 * TH),
					name: n,
				}));
			}
		}
		// fastmath test
		{
			struct FastMath {
				pos: (u16, u16),
				name: &'static [u8],
			}
			impl TestSceneField for FastMath {
				fn draw(&self, ts: &TestScene, active: bool) {
					let pval = ts.deg as u16;
					draw_ascii_text_cached(self.pos.0, self.pos.1, TSX, TSY, &self.name);
					if active {
						fillrect(0, 192, 0, self.pos.0 + (7 + ts.parx as u16) * TW, self.pos.1, TW, TH);
					}
					draw_ascii_text_cached(self.pos.0 +  7 * TW, self.pos.1, TSX, TSY, &format!("{:04}", pval).bytes().collect::<Vec<u8>>());

					use fastmath::{sin, cos, atan2, Q, Qsca, Qang};
					let x = cos(Qang::from_deg(ts.deg)).muli(40);
					let y = sin(Qang::from_deg(ts.deg)).muli(40);
					fillrect(0, 0, 192, (160 + x) as u16, (120 + y) as u16, TW, TH);
					let a2 = atan2(Qsca::from_i32(y), Qsca::from_i32(x)).raw();
					draw_ascii_text_cached(self.pos.0 +  0 * TW, self.pos.1 + TH, TSX, TSY, &format!("{:08X}", a2).bytes().collect::<Vec<u8>>());
				}
				fn input(&mut self, pts: *mut TestScene, btns: &sys::DigitalPadButtonState) {
					let ts = unsafe { pts.as_mut().unwrap() };
					let mut change = |f: &Fn(i32, i32) -> i32| {
						ts.deg = f(ts.deg, 10i32.pow(3 - ts.parx as u32));
					};
					if btns.down().up() {
						change(&|v,d|v+d);
					}
					if btns.down().down() {
						change(&|v,d|v-d);
					}
					if btns.press().triangle() {
						change(&|v,d|v+d); // incr while pressing (per frame)
					}
				}
				fn columns(&self) -> usize {
					4
				}
			}
			fields.push(Box::new(FastMath {
				pos: (par3xy.0, par3xy.1),
				name: &b"deg"[..],
			}));
		}

		// GraphicLogger already caches... but do this anyway.
		cache_ascii_font(0xFFFF);

		sys::init_card(true);
		logger.writeln(b"init_card OK");
		sys::start_card();
		logger.writeln(b"start_card OK");
		sys::bu_init();
		logger.writeln(b"_bu_init OK");

		let (gp1cmd, x1, x2, y1, y2) = calc_gpu_res_params(1, 0, 0, 0);
		gp1(gp1cmd);

		logger.writeln(b"loading JUMP.VB");
		let test1vb = binread(b"cdrom:\\jump.vb;1\0").unwrap();
		spu_dmaload(0x60000, &test1vb);
		logger.writeln(b"JUMP.VB ready");

		TestScene {
			fields: fields,
			gpupars: [x1, x2, y1, y2],
			..Default::default()
		}
	}

#[cfg_attr(not(release), inline(never))]
	fn load_cdfs(&mut self) {
		let cdfs = sys::DirFinder::new(b"cdrom:\\*\0").map(|e| e.name().to_owned() ).collect();
		let cderr = sys::get_last_error();
		self.romfiles = Option::Some((cdfs, cderr));
	}
#[cfg_attr(not(release), inline(never))]
	fn load_bufs(&mut self) {
		let bufs = sys::DirFinder::new(b"bu10:\0").map(|e| e.name().to_owned() ).collect();
		let buerr = sys::get_last_error();
		self.bufiles = Option::Some((bufs, buerr));
	}
#[cfg_attr(not(release), inline(never))]
	fn load_cnf(&mut self) {
		self.cnf = Option::Some(sys::file_open(b"cdrom:\\SYSTEM.CNF;1\0", sys::ReadWrite::Read, sys::file_mode::SYNC).and_then(|mut fd| {
			let mut buf = [0u8; 2048];
			sys::file_read(&mut fd, &mut buf).and_then(|sz| {
				Result::Ok(buf[0..sz].to_vec())
			})
		}));
	}
}

impl Scene for TestScene {
#[cfg_attr(not(release), inline(never))]
	fn draw(&self) {
		gp0(0xE1000000); // texpage (almost for non-texturedpolys)
		gp0(0xE2000000); // texwindow
		gp0(0xE6000000); // maskbits

		let curyres = (((gpustat() >> 19) & 1) + 1) * 240;
		if curyres == 240 {
			// 240p: double-buffering possible
			// switch displaying frame
			let preyoff = self.frame * 256;
			gp1(0x05000000 | (  0 << 0) | (preyoff << 10));
			// switch drawing frame
			let curyoff = (1 - self.frame) * 256;
			gp0(0xE3000000 | (  0 << 0) | (curyoff << 10)); // draw(clip)x1y1
			gp0(0xE4000000 | (640 << 0) | ((curyoff + 240) << 10)); // draw(clip)x2y2
			gp0(0xE5000000 | (  0 << 0) | (curyoff << 11)); // drawoffset
			// fillvram is not affected by drawoffset...
			fillvram(64, 64, 64, 0, curyoff as u16, 640, 240);
		} else {
			// 480i: can't double-buffering
			// fix display frame and drawing frame
			gp1(0x05000000 | (  0 << 0) | (  0 << 10));
			gp0(0xE3000000 | (  0 << 0) | (  0 << 10)); // draw(clip)x1y1
			gp0(0xE4000000 | (640 << 0) | (480 << 10)); // draw(clip)x2y2
			gp0(0xE5000000 | (  0 << 0) | (  0 << 11)); // drawoffset
			fillvram(64, 64, 64, 0, 0, 640, 480);
		}

		/*
		let hello = b"Hello...".iter().map(|c|*c as u16).collect::<Vec<u16>>();
		draw_text_slow(100, 100, 0xFFFF, 0x10000, 0x10000, &hello);
		*/

		let tsx = 0x10000;
		let tsy = 0x10000;
		let tw = (CACHED_FONT_CW as u32 * tsx / 0x10000) as u16;
		let th = (CACHED_FONT_CH as u32 * tsy / 0x10000) as u16;

		let btnxy = (16, 32);
		let btnspw = 8;
		let btnrcwh = (6, 10);
		let cdflxy = (16, 32 + th * 16);
		let buflxy = (16, 32 + th * 21);
		let confxy = (16 + tw * 15, 32 + th * 16);

		// button state
		for i in 0..16 {
			let b = (self.btns.press().bits() >> i) & 1;
			fillrect((128 + b * 127) as u8, (b * 255) as u8, (b * 255) as u8, btnxy.0 + i * btnspw, btnxy.1, btnrcwh.0, btnrcwh.1);
		}

		for (i,f) in self.fields.iter().enumerate() {
			f.draw(self, i == self.pary);
		}

		// debug: cd filelist
		match self.romfiles {
			Some((ref fs, err)) => {
				for (i,n) in fs.iter().enumerate() {
					draw_ascii_text_cached(cdflxy.0, cdflxy.1 + i as u16 * th, tsx, tsy, &n);
				}
				draw_ascii_text_cached(cdflxy.0, cdflxy.1 + fs.len() as u16 * th, tsx, tsy, &format!("{:04X}", err as u16).bytes().collect::<Vec<u8>>());
			},
			None => {
				draw_ascii_text_cached(cdflxy.0, cdflxy.1, tsx, tsy, b"not loaded cd files yet");
			}
		}
		// debug: bu filelist
		match self.bufiles {
			Some((ref fs, err)) => {
				for (i,n) in fs.iter().enumerate() {
					draw_ascii_text_cached(buflxy.0, buflxy.1 + i as u16 * th, tsx, tsy, &n);
				}
				draw_ascii_text_cached(buflxy.0, buflxy.1 + fs.len() as u16 * th, tsx, tsy, &format!("{:04X}", err as u16).bytes().collect::<Vec<u8>>());
			},
			None => {
				draw_ascii_text_cached(buflxy.0, buflxy.1, tsx, tsy, b"not loaded bu files yet");
			}
		}
		// debug: system.cnf content
		match self.cnf {
			Some(Ok(ref fc)) => {
				for (i,ln) in fc.split(|c| *c == b'\n').enumerate() {
					draw_ascii_text_cached(confxy.0, confxy.1 + i as u16 * th, tsx, tsy, &ln);
				}
			},
			Some(Err(err)) => {
				draw_ascii_text_cached(confxy.0, confxy.1, tsx, tsy, &format!("{:04X}", err as u16).bytes().collect::<Vec<u8>>());
			},
			None => {
				draw_ascii_text_cached(confxy.0, confxy.1, tsx, tsy, b"not loaded conf content yet");
			}
		}
	}

#[cfg_attr(not(release), inline(never))]
	fn input(&mut self, bitsu16: u16) {
		self.btns.update(bitsu16);

		let pressbits = self.btns.press();
		let downbits = self.btns.down();

		if downbits.circle() {
			spu_play(0, 0x60000, 22050);
		}

		if downbits.r1() {
			self.pary += 1;
			if self.fields.len() <= self.pary {
				self.pary = 0;
			}
			if self.fields[self.pary].columns() <= self.parx {
				self.parx = self.fields[self.pary].columns() - 1;
			}
		}
		if downbits.l1() {
			if 0 < self.pary {
				self.pary -= 1;
			} else {
				self.pary = self.fields.len() - 1;
			}
			if self.fields[self.pary].columns() <= self.parx {
				self.parx = self.fields[self.pary].columns() - 1;
			}
		}
		if pressbits.cross() {
			if downbits.left() {
				match self.romfiles {
					None => self.load_cdfs(),
					Some(_) => self.romfiles = Option::None,
				}
			}
			if downbits.down() {
				match self.bufiles {
					None => self.load_bufs(),
					Some(_) => self.bufiles = Option::None,
				}
			}
			if downbits.right() {
				match self.cnf {
					None => self.load_cnf(),
					Some(_) => self.cnf = Option::None,
				}
			}
		} else if pressbits.square() {
			if downbits.left() {
				let title = b"Test42".iter().map(|c|ascii2sjis(*c)).collect::<Vec<u16>>();
				let icon = (&[0x44;16], &[[0x22;128]][..]);
				let r = sys::write_bu(b"bu10:BISCPS-00000test\0", sys::file_mode::SYNC, 1, &title, icon, |_fd| {
					Result::Ok(())
				});

				self.wrerr = Option::Some(match r {
					Ok(_) => 0x1212,
					Err(err) => err as u32
				});
			}
			if downbits.right() {
				// debug horz res chg
				let curxres = (gpustat() >> 17) & 3;
				let curyres = (gpustat() >> 19) & 1;
				let currate = (gpustat() >> 20) & 1;
				let curbpp = (gpustat() >> 21) & 1;
				let nextxres = (curxres + 1) & 3;
				let (gp1cmd, x1, x2, y1, y2) = calc_gpu_res_params(nextxres, curyres, curbpp, currate);
				gp1(gp1cmd);
				self.gpupars = [x1, x2, y1, y2];
			}
			if downbits.down() {
				// debug vert res chg
				let curxres = (gpustat() >> 17) & 3;
				let curyres = (gpustat() >> 19) & 1;
				let currate = (gpustat() >> 20) & 1;
				let curbpp = (gpustat() >> 21) & 1;
				let nextyres = (curyres + 1) & 1;
				let (gp1cmd, x1, x2, y1, y2) = calc_gpu_res_params(curxres, nextyres, curbpp, currate);
				gp1(gp1cmd);
				self.gpupars = [x1, x2, y1, y2];
			}
			if downbits.up() {
				// Space Echo
				let params: [u16; 32] = [
					0x033D,0x0231,0x7E00,0x5000,0xB400,0xB000,0x4C00,0xB000,
					0x6000,0x5400,0x1ED6,0x1A31,0x1D14,0x183B,0x1BC2,0x16B2,
					0x1A32,0x15EF,0x15EE,0x1055,0x1334,0x0F2D,0x11F6,0x0C5D,
					0x1056,0x0AE1,0x0AE0,0x07A2,0x0464,0x0232,0x8000,0x8000
				];
				unsafe {
					volatile_store(0x1F801DA2 as *mut u16, 0xE128);
					for (i,p) in params.iter().enumerate() {
						volatile_store((0x1F801DC0 + i * 2) as *mut u16, *p);
					}
				}
			}
		} else {
			if downbits.right() {
				self.parx += 1;
				if self.fields[self.pary].columns() <= self.parx {
					self.parx = 0;
				}
			}
			if downbits.left() {
				if 0 < self.parx {
					self.parx -= 1;
				} else {
					self.parx = self.fields[self.pary].columns() - 1;
				}
			}
		}
		{
			let pts = self as *mut TestScene;
			self.fields[self.pary].input(pts, &self.btns);
		}

		// draw is immutable ref... modify this here.
		self.frame = (self.frame + 1) % 2;
	}

}
